#!/usr/bin/env python 
# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

#
# This file contains the main interface to the QPalma pipeline.
#

import os
import os.path
import pdb
import sys

from QPalma.DatasetUtils import generateTrainingDataset
from QPalma.SettingsParser import parseSettings
from QPalma.sequence_utils import SeqSpliceInfo,DataAccessWrapper
from QPalma.qpalma_main import QPalma
#from QPalma.PipelineHeuristic import PHeuristic

Errormsg = """Usage is:  
                  
                  qpalma_train <config filename> <training data filename> <result dir>                  
           """

def printMessage(mes):
   print '#'*80
   print '\t\t\t%s...\n'%mes
   print '#'*80


class System:
   """
   This class wraps the outer loop of the qpalma project

   It is responsible for:

   - loading and checking the config file(s)
   - setting up the different pipeline modules
   - run the experiment and report the results

   """

   def __init__(self,filename):
      """
      Inititalize the system by loading and parsing the self.settings file to obtain
      all parameters.
      """

      self.settings = parseSettings(filename)
      printMessage('Parsed self.settings system set up.')


   def training(self, training_data_fn, result_dir):
      """
      This function is responsible for the whole training process. It first
      converts the data to the right format needed by QPalma for the training
      algorithm.
      """
      printMessage('Begin of training.')
      self.settings['anzpath']=2 ;
      

      accessWrapper = DataAccessWrapper(self.settings)
      seqInfo = SeqSpliceInfo(accessWrapper, self.settings['allowed_fragments'])

      # Collect the data and create a pickled training set
      training_data_pickle_fn = generateTrainingDataset(training_data_fn,result_dir,self.settings,seqInfo)

      qp = QPalma(seqInfo)
      qp.init_training(training_data_pickle_fn, result_dir, self.settings)

      printMessage('End of training.')


   
if __name__ == '__main__':
   if len(sys.argv) != 5:
      print 'Error: Invalid number of arguments!'
      #print len(sys.argv)
      print Errormsg
      sys.exit(1)

   mode = sys.argv[1]
   if not mode in ['qpalma_train']:
      print Errormsg
      sys.exit(1)

   settings_fn = sys.argv[2]
   if not os.path.exists(settings_fn):
      print 'Error: File %s does not exist' % settings_fn 
      sys.exit(1)

   input_data_fn = sys.argv[3]
   if not os.path.exists(input_data_fn):
      print 'Error: File %s does not exist' % input_data_fn 
      sys.exit(1)

   res_dir = sys.argv[4]
   if not os.path.exists(res_dir):
      print 'Error: Result directory %s does not exist' % res_dir
      sys.exit(1)


   # creating system object
   system_obj = System(settings_fn)

   if mode == 'qpalma_train':
       system_obj.training(input_data_fn,res_dir)
   else:
      assert False
