#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "common.h"
#include "genome.h"

//#define ALLOW_ATAC_INTRONS

const int max_exon_space          = 200000 ;
const int max_intron_space        = 250000 ;

int min(int a, int b)
{
  if (a>b) return b ;
  return a ;
} 

int max(int a, int b)
{
  if (a>b) return a ;
  return b ;
} 

void copy_int(int *dst, int *src, int count)
{
	int i ;
	if (src>dst)
		for (i=0; i<count; i++)
			dst[i]=src[i] ;
	else
		for (i=count-1; i>=0; i--)
			dst[i]=src[i] ;
}

void copy_float(float *dst, float *src, int count)
{
	int i ;
	if (src>dst)
		for (i=0; i<count; i++)
			dst[i]=src[i] ;
	else
		for (i=count-1; i>=0; i--)
			dst[i]=src[i] ;
}

void reverse_array(int * ar, int length)
{
  int * buf = (int *)malloc(sizeof(int)*length) ;
  int i ;
  for (i=0; i<length; i++)
    buf[length-i-1]=ar[i] ;
  copy_int(ar, buf, length) ;
  free(buf) ;
} ;

void reverse_array_float(float * ar, int length)
{
  float * buf = (float *)malloc(sizeof(float)*length) ;
  int i ;
  for (i=0; i<length; i++)
    buf[length-i-1]=ar[i] ;
  copy_float(ar, buf, length) ;
  free(buf) ;
} ;

/* read a line from a file */
char *read_line(FILE*infile, char* line, size_t max_len) 
{
	int i,c ;
	
	i=0 ;
	while (!feof(infile))
    {
		c=fgetc(infile) ;
		if ((c==EOF) || (c==13) || (c==10))
			break ;
		line[i]=c ;
		
		i++ ;
		if (i>max_len)
		{
			fprintf(stderr, "line buffer not large enough\n") ;
			exit(-1) ;
		} ;
    } ;
	line[i]=0 ;
	
	if (i==0)
		strcpy(line,"") ;
	return line ;
}

/* reverse complement a sequence */
char *reverse_complement(char*in)
{
  char *out=strdup(in) ;
  int i, len=strlen(in) ;
  for (i=0; i<len; i++)
  {
      switch (in[i])
	  {
	  case 'a': out[len-i-1]='t' ; break ;
	  case 't': out[len-i-1]='a' ; break ;
	  case 'g': out[len-i-1]='c' ; break ;
	  case 'c': out[len-i-1]='g' ; break ;
	  case 'n': out[len-i-1]='n' ; break ;
	  case 'A': out[len-i-1]='T' ; break ;
	  case 'T': out[len-i-1]='A' ; break ;
	  case 'G': out[len-i-1]='C' ; break ;
	  case 'C': out[len-i-1]='G' ; break ;
	  case 'N': out[len-i-1]='N' ; break ;
	  case '-': out[len-i-1]='-' ; break ;
	  default:
		  if (in[i]>='a' && in[i]<='z')
		  {
			  fprintf(stderr,"wrong letter %c -> reverse complementing to 'n'\n", in[i]) ;
			  continue ;
		  }
		  if (in[i]>='A' && in[i]<='Z')
		  {
			  fprintf(stderr,"wrong letter %c -> reverse complementing to 'N'\n", in[i]) ;
			  continue ;
		  }
		  fprintf(stderr,"wrong letter '%c' -> aborting", in[i]) ;
		  exit(-1)  ;
	  } ;
  } ;
  return out ;
} 

/* find the maximal orf from left to right in one frame*/
int find_open_frame(char * seq, int len)
{
  int i;
  for (i=0; i<len-2; i+=3)
    if (seq[i]==alph[3]) {
      if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[0])) break ;
      if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[2])) break ;
      if ((seq[i+1]==alph[2]) && (seq[i+2]==alph[0])) break ;
    } ;
  /*  int offset=0 ;
  if (seq[i]==alph[3]) {
    if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[0])) offset=3 ;
    if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[2])) offset=3 ;
    if ((seq[i+1]==alph[2]) && (seq[i+2]==alph[0])) offset=3 ;
  } ;
  i-=offset ;*/
  if (i>=len-2)
    i=len ;
//  if (verbosity>1)
//    fprintf(stdout, "%s %i %i\n", &seq[i], i, len) ;
  return i;    
} 

/* find the maximal orf from left to right in all frames */
int find_max_open_frame(char * seq, int len,
			int *pl1, int *pl2, int *pl3)
{
  int l1=find_open_frame(seq, len) ;
  int l2=find_open_frame(&(seq[1]), len-1)+1 ;
  int l3=find_open_frame(&(seq[2]), len-2)+2 ;
  if (pl1) *pl1=l1 ;
  if (pl2) *pl2=l2 ;
  if (pl3) *pl3=l3 ;
  if ((l1>l2) && (l1>l3)) return min(l1,len) ;
  if (l2>l3) return min(l2,len) ;
  return min(l3,len) ;
}

/* find the maximal orf from right to left in one frame */
int find_open_frame_backward(char * seq, int len)
{
  int i ;

  for (i=len-3; i>=0; i-=3) 
    if (seq[i]==alph[3]) {
      if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[0])) break ;
      if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[2])) break ;
      if ((seq[i+1]==alph[2]) && (seq[i+2]==alph[0])) break ;
    } ;
  int offset=0 ;
  if ((i>=0) &&(seq[i]==alph[3]) ) {
    if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[0])) offset=3 ;
    if ((seq[i+1]==alph[0]) && (seq[i+2]==alph[2])) offset=3 ;
    if ((seq[i+1]==alph[2]) && (seq[i+2]==alph[0])) offset=3 ;
  } ;
  i+=offset ;
  if (i<0)
    i=0 ;
//  if (verbosity>1)
//    fprintf(stdout, "%s %i %i [%i]\n", &seq[i], len-i, len, strlen(&seq[i])) ;
  return len-i;    
} 

/* find the maximal orf from right to left in all frames */
int find_max_open_frame_backward(char * seq, int len, 
				 int *pl1, int *pl2, int *pl3)
{
  int l1=find_open_frame_backward(seq, len) ;
  int l2=find_open_frame_backward(seq, len-1)+1 ;
  int l3=find_open_frame_backward(seq, len-2)+2 ;
  if (pl1) *pl1=l1 ;
  if (pl2) *pl2=l2 ;
  if (pl3) *pl3=l3 ;
  if ((l1>l2) && (l1>l3)) return min(l1,len) ;
  if (l2>l3) return min(l2,len) ;
  return min(l3,len) ;
}

/* checks exons for stop codons at inappropriate positions */
int get_exon(int contig_num, int start, int end, char ori, 
	     char**pexon) 
{
  int ok=1 ;
  char * exon=NULL ;
  int exon_len=end-start+1 ;
  
  if (exon_len+1>max_exon_space) 
  {
      fprintf(stderr, "exon buffer too small\n") ;
      exit(-1) ;
  } ;
  if (ori=='+')
  {
      exon=(char*)malloc(max_exon_space*sizeof(char)) ;
      if (fseek(contig_files[contig_num], start-1, SEEK_SET))
	  {
		  fprintf(stderr, "fseek error (1a) %i %i\n", contig_num, start-1) ;
		  exit(-1) ;
	  } ;
      if (fread(exon, sizeof(char), exon_len, contig_files[contig_num])!=exon_len)
	  {
		  fprintf(stderr, "fread error (1a) %i %i %i\n", contig_num, exon_len, start-1) ;
		  exit(-1) ;
	  } ;
      exon[exon_len]=0 ;
  } 
  else
	  if (ori=='-')
      {
		  char * exon_=(char*)malloc(max_exon_space*sizeof(char)) ;
		  int read ; 
		  
		  if (fseek(contig_files[contig_num], start-1, SEEK_SET))
		  {
			  fprintf(stderr, "fseek error (2a) %i %i\n", contig_num, start-1) ;
			  exit(-1) ;
		  } ;
		  if ((read=fread(exon_, sizeof(char), exon_len, contig_files[contig_num]))!=exon_len)
		  {
			  fprintf(stderr, "fread error (2a) %i %i %i %i\n", contig_num, exon_len, read, start-1) ;
			  exit(-1) ;
		  } ;
		  
		  exon_[exon_len]=0 ;
		  exon=reverse_complement(exon_) ;
		  free(exon_) ;
      } 
	  else {
		  fprintf(stderr, "wrong orientation %c\n",ori) ;
		  exit(-1) ;
	  } ;
  
  /* check exon */
  *pexon=exon ;
  
  return ok ;
} 


/* checks exons for stop codons at inappropriate positions */
int check_exon(int contig_num, int start, int end, char ori, 
			   char**pexon, int *porf_len_left, int *porf_len_right) 
{
	int ok=1 ;
	char * exon=NULL ;
	int orf_len_left=0 ;
	int orf_len_right=0 ;
	int exon_len=end-start+1 ;
	//  fprintf(stderr, "check_exon: start=%i end=%i\n", start, end) ;
	
	if (exon_len+1>max_exon_space) 
    {
		fprintf(stderr, "exon buffer too small A [%i]\n", exon_len+1) ;
		exit(-1) ;
    } ;
	if (ori=='+')
    {
		exon=(char*)malloc(max_exon_space*sizeof(char)) ;
		if (fseek(contig_files[contig_num], start-1, SEEK_SET))
		{
			fprintf(stderr, "fseek error (1b) %i %i\n", contig_num, start-1) ;
			exit(-1) ;
		} ;
		if (fread(exon, sizeof(char), exon_len, contig_files[contig_num])!=exon_len)
		{
			fprintf(stderr, "fread error (1b) %i %i %i\n", contig_num, exon_len, start-1) ;
			exit(-1) ;
		} ;
		exon[exon_len]=0 ;
    } 
	else
		if (ori=='-')
		{
			char * exon_=(char*)malloc(max_exon_space*sizeof(char)) ;
			int read ; 
			
			if (fseek(contig_files[contig_num], start-1, SEEK_SET))
			{
				fprintf(stderr, "fseek error (2b) %i %i\n", contig_num, start-1) ;
				exit(-1) ;
			} ;
			if ((read=fread(exon_, sizeof(char), exon_len, contig_files[contig_num]))!=exon_len)
			{
				fprintf(stderr, "fread error (2b) %i %i %i %i\n", contig_num, exon_len, read, start-1) ;
				exit(-1) ;
			} ;
			
			exon_[exon_len]=0 ;
			exon=reverse_complement(exon_) ;
			free(exon_) ;
		} 
		else {
			fprintf(stderr, "wrong orientation %c\n",ori) ;
			exit(-1) ;
		} ;
	
	/* check exon */
	*porf_len_left=-1 ;
	*porf_len_right=-1 ;
	orf_len_left=find_max_open_frame(exon, exon_len,0,0,0) ;
	orf_len_right=find_max_open_frame_backward(exon, exon_len,0,0,0) ;
	/*  fprintf(stdout, "%i\t %i, %i\n", start, orf_len, exon_len) ; */
	if (orf_len_left < max(exon_len-3,0))
    {
		ok=0 ;
		*pexon=exon ;
		*porf_len_left=orf_len_left ;
		*porf_len_right=orf_len_right ;
    }
	else
		if (*pexon != (char*)-1)
		{
			*pexon=NULL ;
			free(exon) ;
		}
		else
			*pexon = exon ;
	
	return ok ;
} 

int check_intron(int contig_num, int start, int end, char ori, 
				 int *start_diff, int *end_diff)
{
	const int max_move = 2 ;
	char * intron=NULL ;
	int intron_len=end-start+1 ;
	
	//fprintf(stdout, "check intron: start=%i, end=%i\n", start, end) ;
	
	//assert(intron_len>=0) ;
	if (intron_len<6)
	{
		fprintf(stdout, "intron_len<6 [%i]\n", intron_len) ;
		return 0 ;
	}
	
	if (intron_len+1+2*max_move>max_intron_space) 
    {
		fprintf(stderr, "intron buffer too small (%i,%i)\n",intron_len,max_intron_space) ;
		exit(-1) ;
    } ;
	if (ori=='+')
    {
		intron=(char*)malloc(max_intron_space*sizeof(char)) ;
		if (fseek(contig_files[contig_num], start-1-max_move, SEEK_SET))
		{
			fprintf(stderr, "fseek error (10) %i %i\n", contig_num, start-1) ;
			exit(-1) ;
		} ;
		if (fread(intron, sizeof(char), intron_len+2*max_move, contig_files[contig_num])!=intron_len+2*max_move)
		{
			fprintf(stderr, "fread error (10) %i %i %i\n", contig_num, intron_len, start-1) ;
			exit(-1) ;
		} ;
		intron[intron_len+2*max_move]=0 ;
    } 
	else
		if (ori=='-')
		{
			char * intron_=(char*)malloc(max_intron_space*sizeof(char)) ;
			int read ; 
			
			if (fseek(contig_files[contig_num], start-1-max_move, SEEK_SET))
			{
				fprintf(stderr, "fseek error (11) %i %i\n", contig_num, start-1) ;
				exit(-1) ;
			} ;
			if ((read=fread(intron_, sizeof(char), intron_len+2*max_move, contig_files[contig_num]))!=intron_len+2*max_move)
			{
				fprintf(stderr, "fread error (11) %i %i %i %i\n", contig_num, intron_len, read, start-1) ;
				exit(-1) ;
			} ;
			
			intron_[intron_len+2*max_move]=0 ;
			intron=reverse_complement(intron_) ;
			free(intron_) ;
		} 
		else {
			fprintf(stderr, "wrong orientation %c\n",ori) ;
			exit(-1) ;
		} ;
	
	int ok= (intron[max_move]==alph[2]) && (intron[max_move+1]==alph[3]) && 
		(intron[intron_len-2+max_move]==alph[0]) && (intron[intron_len-1+max_move]==alph[2])  ;
	ok= ok || ((intron[max_move]==alph[2]) && (intron[max_move+1]==alph[1]) && 
		(intron[intron_len-2+max_move]==alph[0]) && (intron[intron_len-1+max_move]==alph[2]))  ;
#ifdef ALLOW_ATAC_INTRONS
	ok= ok || ((intron[max_move]==alph[0]) && (intron[max_move+1]==alph[3]) && 
		(intron[intron_len-2+max_move]==alph[0]) && (intron[intron_len-1+max_move]==alph[1]))  ;
#endif

	*start_diff=0 ;
	*end_diff=0 ;
	if (!ok)
    {
		if ((max_move>=1) && (intron[max_move-1]==alph[2]) && 
			(intron[max_move]==alph[3]))
			*start_diff=-1 ;
		if ((max_move>=2) && (intron[max_move-2]==alph[2]) && 
			(intron[max_move-1]==alph[3]))
			*start_diff=-2 ;
		if ((max_move>=1) && (intron[max_move+1]==alph[2]) && 
			(intron[max_move+2]==alph[3]))
			*start_diff= 1 ;
		if ((max_move>=2) && (intron[max_move+2]==alph[2]) && 
			(intron[max_move+3]==alph[3]))
			*start_diff= 2 ;
		if ((max_move>=1) && (intron[intron_len-2+max_move-1]==alph[0]) && 
			(intron[intron_len-1+max_move-1]==alph[2]))
			*end_diff=-1 ;
		if ((max_move>=2) && (intron[intron_len-2+max_move-2]==alph[0]) && 
			(intron[intron_len-1+max_move-2]==alph[2]))
			*end_diff=-2 ;
		if ((max_move>=1) && (intron[intron_len-2+max_move+1]==alph[0]) && 
			(intron[intron_len-1+max_move+1]==alph[2]))
			*end_diff= 1 ;
		if ((max_move>=2) && (intron[intron_len-2+max_move+2]==alph[0]) && 
			(intron[intron_len-1+max_move+2]==alph[2]))
			*end_diff= 2 ;
		
		ok= (intron[max_move + *start_diff]==alph[2]) && 
			(intron[max_move+1 + *start_diff]==alph[3]) && 
			(intron[intron_len-2+max_move+*end_diff]==alph[0]) && 
			(intron[intron_len-1+max_move+*end_diff]==alph[2])  ;

		if (!ok)
		{
			*start_diff=0 ;

			if ((max_move>=1) && (intron[max_move-1]==alph[2]) && 
				(intron[max_move]==alph[1]))
				*start_diff=-1 ;
			if ((max_move>=2) && (intron[max_move-2]==alph[2]) && 
				(intron[max_move-1]==alph[1]))
				*start_diff=-2 ;
			if ((max_move>=1) && (intron[max_move+1]==alph[2]) && 
				(intron[max_move+2]==alph[1]))
				*start_diff= 1 ;
			if ((max_move>=2) && (intron[max_move+2]==alph[2]) && 
				(intron[max_move+3]==alph[1]))
				*start_diff= 2 ;
			
			ok= (intron[max_move + *start_diff]==alph[2]) && 
				(intron[max_move+1 + *start_diff]==alph[1]) && 
				(intron[intron_len-2+max_move+*end_diff]==alph[0]) && 
				(intron[intron_len-1+max_move+*end_diff]==alph[2])  ;
		}

#ifdef ALLOW_ATAC_INTRONS
		if (!ok)
		{
			*start_diff=0 ;

			if ((max_move>=1) && (intron[max_move-1]==alph[0]) && 
				(intron[max_move]==alph[3]))
				*start_diff=-1 ;
			if ((max_move>=2) && (intron[max_move-2]==alph[0]) && 
				(intron[max_move-1]==alph[3]))
				*start_diff=-2 ;
			if ((max_move>=1) && (intron[max_move+1]==alph[0]) && 
				(intron[max_move+2]==alph[3]))
				*start_diff= 1 ;
			if ((max_move>=2) && (intron[max_move+2]==alph[0]) && 
				(intron[max_move+3]==alph[3]))
				*start_diff= 2 ;
			
			ok= (intron[max_move + *start_diff]==alph[0]) && 
				(intron[max_move+1 + *start_diff]==alph[3]) && 
				(intron[intron_len-2+max_move+*end_diff]==alph[0]) && 
				(intron[intron_len-1+max_move+*end_diff]==alph[1])  ;
		}
#endif		
		if (ok) ok=-1 ;
		
		fprintf(stdout,"### intron %c%c%c%c|%c%c|%c%c%c%c..%c%c%c%c|%c%c|%c%c  %i %i  [%i]\n", intron[0],intron[1], 
				intron[0], intron[1], intron[2], intron[3],
				intron[4], intron[5], intron[6], intron[7],
				intron[intron_len+2*max_move-8], intron[intron_len+2*max_move-7],
				intron[intron_len+2*max_move-6], intron[intron_len+2*max_move-5],
				intron[intron_len+2*max_move-4], intron[intron_len+2*max_move-3],
				intron[intron_len+2*max_move-2], intron[intron_len+2*max_move-1],
				*start_diff, *end_diff, ok) ;
    } ;
	free(intron) ;
	return ok ;
} 

int check_intron_protein(int contig_num, int start, int end, char ori, 
						 int *start_diff, int *end_diff, int Qgap)
{
	const int max_move = 5 ;
	char * intron=NULL ;
	int intron_len=end-start+1 ;
	
	//assert(intron_len>=0) ;
	if (intron_len<6)
	{
		fprintf(stdout, "intron_len<6 [%i]\n", intron_len) ;
		return 0 ;
	}
	
	if (intron_len+1+2*max_move>max_intron_space) 
    {
		fprintf(stderr, "intron buffer too small (%i,%i)\n",intron_len,max_intron_space) ;
		exit(-1) ;
    } ;
	if (ori=='+')
    {
		intron=(char*)malloc(max_intron_space*sizeof(char)) ;
		if (fseek(contig_files[contig_num], start-1-max_move, SEEK_SET))
		{
			fprintf(stderr, "fseek error (10) %i %i\n", contig_num, start-1) ;
			exit(-1) ;
		} ;
		if (fread(intron, sizeof(char), intron_len+2*max_move, contig_files[contig_num])!=intron_len+2*max_move)
		{
			fprintf(stderr, "fread error (10) %i %i %i\n", contig_num, intron_len, start-1) ;
			exit(-1) ;
		} ;
		intron[intron_len+2*max_move]=0 ;
    } 
	else
		if (ori=='-')
		{
			char * intron_=(char*)malloc(max_intron_space*sizeof(char)) ;
			int read ; 
			
			if (fseek(contig_files[contig_num], start-1-max_move, SEEK_SET))
			{
				fprintf(stderr, "fseek error (11) %i %i\n", contig_num, start-1) ;
				exit(-1) ;
			} ;
			if ((read=fread(intron_, sizeof(char), intron_len+2*max_move, contig_files[contig_num]))!=intron_len+2*max_move)
			{
				fprintf(stderr, "fread error (11) %i %i %i %i\n", contig_num, intron_len, read, start-1) ;
				exit(-1) ;
			} ;
			
			intron_[intron_len+2*max_move]=0 ;
			intron=reverse_complement(intron_) ;
			free(intron_) ;
		} 
		else {
			fprintf(stderr, "wrong orientation %c\n",ori) ;
			exit(-1) ;
		} ;
	
	int ok= (intron[max_move]==alph[2]) && (intron[max_move+1]==alph[3]) && 
		(intron[intron_len-2+max_move]==alph[0]) && (intron[intron_len-1+max_move]==alph[2])  ;
	ok= ok || ((intron[max_move]==alph[2]) && (intron[max_move+1]==alph[1]) && 
		(intron[intron_len-2+max_move]==alph[0]) && (intron[intron_len-1+max_move]==alph[2]))  ;

	*start_diff=0 ;
	*end_diff=0 ;
	if (!ok)
    {
		int move_start = 0, move_end=0, stop_it=0 ;

		for (move_start=1; (move_start<=max_move) && (!stop_it); )
		{
			for (move_end=1; move_end<=max_move; )
			{
				if (1)//((Qgap==1) && (move_start-move_end==3)) || (move_start==move_end))
				{
					if ((intron[max_move+move_start]==alph[2]) && ( (intron[max_move+1+move_start]==alph[3]) || (intron[max_move+1+move_start]==alph[1]) ) &&
						(intron[intron_len-2+max_move+move_end]==alph[0]) &&  (intron[intron_len-1+max_move+move_end]==alph[2]) )
					{
						*start_diff=move_start ;
						*end_diff=move_end ;
						//fprintf(stderr, "match found: %i %i\n", move_start, move_end) ;
						stop_it=1 ;
						break ;
					}
				}
				
				if (move_end>0)
					move_end *= -1 ;
				else
					move_end = -move_end+1 ;
			}
			if (move_start>0)
				move_start *=-1 ;
			else
				move_start = -move_start+1 ;
		}
		
		/*fprintf(stderr, "%i, %i %i, %i\n", (intron[max_move + *start_diff]==alph[2]), 
				((intron[max_move+1 + *start_diff]==alph[3])||(intron[max_move+1 + *start_diff]==alph[1])),
				(intron[intron_len-2+max_move+*end_diff]==alph[0]), (intron[intron_len-1+max_move+*end_diff]==alph[2]));*/

		ok= (intron[max_move + *start_diff]==alph[2]) && 
			((intron[max_move+1 + *start_diff]==alph[3])||(intron[max_move+1 + *start_diff]==alph[1])) &&
			(intron[intron_len-2+max_move+*end_diff]==alph[0]) && 
			(intron[intron_len-1+max_move+*end_diff]==alph[2])  ;
		
		if (ok) ok=-1 ;
		
		fprintf(stdout,"### protein_intron %c%c%c%c|%c%c|%c%c%c%c..%c%c%c%c|%c%c|%c%c  %i %i  [%i]  Qgap=%i\n", intron[0],intron[1], 
				intron[0], intron[1], intron[2], intron[3],
				intron[4], intron[5], intron[6], intron[7],
				intron[intron_len+2*max_move-8], intron[intron_len+2*max_move-7],
				intron[intron_len+2*max_move-6], intron[intron_len+2*max_move-5],
				intron[intron_len+2*max_move-4], intron[intron_len+2*max_move-3],
				intron[intron_len+2*max_move-2], intron[intron_len+2*max_move-1],
				*start_diff, *end_diff, ok, Qgap) ;
    } ;
	free(intron) ;
	return ok ;
} 

int check_intron_revcomp(int contig_num, int start, int end, char ori, 
				 int *start_diff, int *end_diff)
{
	const int max_move = 2 ;
	char * intron=NULL ;
	int intron_len=end-start+1 ;
	
	//fprintf(stdout, "check intron: start=%i, end=%i\n", start, end) ;
	
	//assert(intron_len>=0) ;
	if (intron_len<6)
	{
		fprintf(stdout, "intron_len<6 [%i]\n", intron_len) ;
		return 0 ;
	}
	
	if (intron_len+1+2*max_move>max_intron_space) 
    {
		fprintf(stderr, "intron buffer too small (%i,%i)\n",intron_len,max_intron_space) ;
		exit(-1) ;
    } ;
	if (ori=='+')
    {
		intron=(char*)malloc(max_intron_space*sizeof(char)) ;
		if (fseek(contig_files[contig_num], start-1-max_move, SEEK_SET))
		{
			fprintf(stderr, "fseek error (10) %i %i\n", contig_num, start-1) ;
			exit(-1) ;
		} ;
		if (fread(intron, sizeof(char), intron_len+2*max_move, contig_files[contig_num])!=intron_len+2*max_move)
		{
			fprintf(stderr, "fread error (10) %i %i %i\n", contig_num, intron_len, start-1) ;
			exit(-1) ;
		} ;
		intron[intron_len+2*max_move]=0 ;
    } 
	else
		if (ori=='-')
		{
			char * intron_=(char*)malloc(max_intron_space*sizeof(char)) ;
			int read ; 
			
			if (fseek(contig_files[contig_num], start-1-max_move, SEEK_SET))
			{
				fprintf(stderr, "fseek error (11) %i %i\n", contig_num, start-1) ;
				exit(-1) ;
			} ;
			if ((read=fread(intron_, sizeof(char), intron_len+2*max_move, contig_files[contig_num]))!=intron_len+2*max_move)
			{
				fprintf(stderr, "fread error (11) %i %i %i %i\n", contig_num, intron_len, read, start-1) ;
				exit(-1) ;
			} ;
			
			intron_[intron_len+2*max_move]=0 ;
			intron=reverse_complement(intron_) ;
			free(intron_) ;
		} 
		else {
			fprintf(stderr, "wrong orientation %c\n",ori) ;
			exit(-1) ;
		} ;
	
	// GT-AG -> CT-AC
	int ok= (intron[max_move]==alph[1]) && (intron[max_move+1]==alph[3]) && 
		(intron[intron_len-2+max_move]==alph[0]) && (intron[intron_len-1+max_move]==alph[1])  ;
	// GT-AG -> CT-GC
	ok= ok || ((intron[max_move]==alph[1]) && (intron[max_move+1]==alph[3]) && 
		(intron[intron_len-2+max_move]==alph[3]) && (intron[intron_len-1+max_move]==alph[1]))  ;
#ifdef ALLOW_ATAC_INTRONS
	// AT-AC -> GT-AT
	ok= ok || ((intron[max_move]==alph[3]) && (intron[max_move+1]==alph[3]) && 
		(intron[intron_len-2+max_move]==alph[0]) && (intron[intron_len-1+max_move]==alph[3]))  ;
#endif
	*start_diff=0 ;
	*end_diff=0 ;
	if (!ok) // only do the correction for the most frequent ones
    {
		if ((max_move>=1) && (intron[max_move-1]==alph[1]) && 
			(intron[max_move]==alph[3]))
			*start_diff=-1 ;
		if ((max_move>=2) && (intron[max_move-2]==alph[1]) && 
			(intron[max_move-1]==alph[3]))
			*start_diff=-2 ;
		if ((max_move>=1) && (intron[max_move+1]==alph[1]) && 
			(intron[max_move+2]==alph[3]))
			*start_diff= 1 ;
		if ((max_move>=2) && (intron[max_move+2]==alph[1]) && 
			(intron[max_move+3]==alph[3]))
			*start_diff= 2 ;
		if ((max_move>=1) && (intron[intron_len-2+max_move-1]==alph[0]) && 
			(intron[intron_len-1+max_move-1]==alph[1]))
			*end_diff=-1 ;
		if ((max_move>=2) && (intron[intron_len-2+max_move-2]==alph[0]) && 
			(intron[intron_len-1+max_move-2]==alph[1]))
			*end_diff=-2 ;
		if ((max_move>=1) && (intron[intron_len-2+max_move+1]==alph[0]) && 
			(intron[intron_len-1+max_move+1]==alph[1]))
			*end_diff= 1 ;
		if ((max_move>=2) && (intron[intron_len-2+max_move+2]==alph[0]) && 
			(intron[intron_len-1+max_move+2]==alph[1]))
			*end_diff= 2 ;
		
		ok= (intron[max_move + *start_diff]==alph[1]) && 
			(intron[max_move+1 + *start_diff]==alph[3]) && 
			(intron[intron_len-2+max_move+*end_diff]==alph[0]) && 
			(intron[intron_len-1+max_move+*end_diff]==alph[1])  ;

		if (ok) ok=-1 ;
    } ;
	free(intron) ;
	return ok ;
} 


/* checks exons for stop codons at inappropriate positions */
int check_intron_simple(int contig_num, int start, int end, char ori)
{
	char * intron=NULL ;
	start-=2 ;
	end+=2 ;
	int intron_len=end-start+1 ;
	
	if (intron_len+1>max_intron_space) 
    {
		fprintf(stderr, "intron buffer too small\n") ;
		exit(-1) ;
    } ;
	if (ori=='+')
    {
		intron=(char*)malloc(max_intron_space*sizeof(char)) ;
		if (fseek(contig_files[contig_num], start-1, SEEK_SET))
		{
			fprintf(stderr, "fseek error (1c) %i %i\n", contig_num, start-1) ;
			exit(-1) ;
		} ;
		if (fread(intron, sizeof(char), intron_len, contig_files[contig_num])!=intron_len)
		{
			fprintf(stderr, "fread error (1c) %i %i %i\n", contig_num, intron_len, start-1) ;
			exit(-1) ;
		} ;
		intron[intron_len]=0 ;
    } 
	else
		if (ori=='-')
		{
			char * intron_=(char*)malloc(max_intron_space*sizeof(char)) ;
			int read ; 
			
			if (fseek(contig_files[contig_num], start-1, SEEK_SET))
			{
				fprintf(stderr, "fseek error (2c) %i %i\n", contig_num, start-1) ;
				exit(-1) ;
			} ;
			if ((read=fread(intron_, sizeof(char), intron_len, contig_files[contig_num]))!=intron_len)
			{
				fprintf(stderr, "fread error (2c) %i %i %i %i\n", contig_num, intron_len, read, start-1) ;
				exit(-1) ;
			} ;
			
			intron_[intron_len]=0 ;
			intron=reverse_complement(intron_) ;
			free(intron_) ;
		} 
		else {
			fprintf(stderr, "wrong orientation %c\n",ori) ;
			exit(-1) ;
		} ;
	
	
	int ok= (intron[2]==alph[2]) && (intron[3]==alph[3]) && (intron[intron_len-2-2]==alph[0]) &&(intron[intron_len-1-2]==alph[2])  ;
	ok= ok || ((intron[2]==alph[2]) && (intron[3]==alph[1]) && (intron[intron_len-2-2]==alph[0]) &&(intron[intron_len-1-2]==alph[2]))  ;
#ifdef ALLOW_ATAC_INTRONS
	ok= ok || ((intron[2]==alph[0]) && (intron[3]==alph[3]) && (intron[intron_len-2-2]==alph[0]) &&(intron[intron_len-1-2]==alph[1]))  ;
#endif	
	if (!ok)
	{
		fprintf(stdout,"### intron %i%c: %c%c%c%c%c%c..%c%c%c%c%c%c\n", contig_num, ori, intron[0],intron[1], intron[2], intron[3], intron[4], intron[5], intron[intron_len-6], intron[intron_len-5], intron[intron_len-4], intron[intron_len-3], intron[intron_len-2], intron[intron_len-1]) ;
	};
	
	free(intron) ;
	return ok ;
} 

int get_intron_type(int contig_num, int start, int end, char ori)
{
	char * intron=NULL ;
	start-=2 ;
	end+=2 ;
	int intron_len=end-start+1 ;
	
	if (intron_len+1>max_intron_space) 
    {
		fprintf(stderr, "intron buffer too small\n") ;
		exit(-1) ;
    } ;
	if (ori=='+')
    {
		intron=(char*)malloc(max_intron_space*sizeof(char)) ;
		if (fseek(contig_files[contig_num], start-1, SEEK_SET))
		{
			fprintf(stderr, "fseek error (1d) %i %i\n", contig_num, start-1) ;
			exit(-1) ;
		} ;
		if (fread(intron, sizeof(char), intron_len, contig_files[contig_num])!=intron_len)
		{
			fprintf(stderr, "fread error (1d) %i %i %i\n", contig_num, intron_len, start-1) ;
			exit(-1) ;
		} ;
		intron[intron_len]=0 ;
    } 
	else
		if (ori=='-')
		{
			char * intron_=(char*)malloc(max_intron_space*sizeof(char)) ;
			int read ; 
			
			if (fseek(contig_files[contig_num], start-1, SEEK_SET))
			{
				fprintf(stderr, "fseek error (2d) %i %i\n", contig_num, start-1) ;
				exit(-1) ;
			} ;
			if ((read=fread(intron_, sizeof(char), intron_len, contig_files[contig_num]))!=intron_len)
			{
				fprintf(stderr, "fread error (2d) %i %i %i %i\n", contig_num, intron_len, read, start-1) ;
				exit(-1) ;
			} ;
			
			intron_[intron_len]=0 ;
			intron=reverse_complement(intron_) ;
			free(intron_) ;
		} 
		else {
			fprintf(stderr, "wrong orientation %c\n",ori) ;
			exit(-1) ;
		} ;
	
	int ok=-1 ;
	
	if ((intron[2]==alph[2]) && (intron[3]==alph[3]) && (intron[intron_len-2-2]==alph[0]) &&(intron[intron_len-1-2]==alph[2]))
		ok=1 ;
	
	if ((intron[2]==alph[2]) && (intron[3]==alph[1]) && (intron[intron_len-2-2]==alph[0]) &&(intron[intron_len-1-2]==alph[2]))
		ok=2 ;

#ifdef ALLOW_ATAC_INTRONS
	if ((intron[2]==alph[0]) && (intron[3]==alph[3]) && (intron[intron_len-2-2]==alph[0]) &&(intron[intron_len-1-2]==alph[1]))
		ok=3 ;
#endif

	free(intron) ;

	return ok ;
} 


int get_intron(int contig_num, int start, int end, char ori, char **intron)
{
	*intron=NULL ;
	int intron_len=end-start+1 ;
	
	if (intron_len+1>max_intron_space) 
    {
		fprintf(stderr, "intron buffer too small\n") ;
		exit(-1) ;
    } ;
	if (ori=='+')
    {
		*intron=(char*)malloc(max_intron_space*sizeof(char)) ;
		if (fseek(contig_files[contig_num], start-1, SEEK_SET))
		{
			(*intron)[0]=0 ;
			fprintf(stderr, "fseek error (1e) %i %i\n", contig_num, start-1) ;
			return 0 ;
		} ;
		if (fread(*intron, sizeof(char), intron_len, contig_files[contig_num])!=intron_len)
		{
			(*intron)[0]=0 ;
			fprintf(stderr, "fread error (1e) %i %i %i\n", contig_num, intron_len, start-1) ;
			return 0 ;
		} ;
		(*intron)[intron_len]=0 ;
    } 
	else
		if (ori=='-')
		{
			char * intron_=(char*)malloc(max_intron_space*sizeof(char)) ;
			int read ; 
			
			if (fseek(contig_files[contig_num], start-1, SEEK_SET))
			{
				*intron = intron_ ;
				(*intron)[0]=0 ;
				fprintf(stderr, "fseek error (2e) %i %i\n", contig_num, start-1) ;
				return 0 ;
			} ;
			if ((read=fread(intron_, sizeof(char), intron_len, contig_files[contig_num]))!=intron_len)
			{
				*intron = intron_ ;
				(*intron)[0]=0 ;
				fprintf(stderr, "fread error (2e) %i %i %i %i\n", contig_num, intron_len, read, start-1) ;
				return 0 ;
			} ;
			
			intron_[intron_len]=0 ;
			*intron=reverse_complement(intron_) ;
			free(intron_) ;
		} 
		else {
			fprintf(stderr, "wrong orientation %c\n",ori) ;
			exit(-1) ;
		} ;
	return 1 ;
} 

