#/bin/bash

set -e

PROG=`basename $0`
DIR=`dirname $0`

. $DIR/../bin/qpalma_config.sh

export data_file=$1
export result_dir=$2
mkdir -p $result_dir

if [ "$QPALMA_genome_fasta" != "" ]
then
    export QPALMA_genome_dir_base=`tempfile`
    rm $QPALMA_genome_dir_base; mkdir $QPALMA_genome_dir_base
    python $QPALMA_PATH/galaxy/make_gio.py $QPALMA_genome_fasta $QPALMA_genome_dir_base
    export QPALMA_genome_dir=$QPALMA_genome_dir_base/genome
fi

#QPALMA_contigs=`seq 1 $num_contigs`
num_contigs=`grep CONTIGS $QPALMA_genome_dir/../genome.config | sed "s/CONTIGS//g"`
export num_contigs
line_num=`cat $QPALMA_genome_dir/../genome.config | grep -n CONTIGS | sed "s/\(.*\):.*/\1/g"`

stop=$((line_num+num_contigs))
start=$((line_num+1))
QPALMA_contigs=""
for i in `seq $start $stop`;
do
    QPALMA_contigs=`sed -n ${i}p $QPALMA_genome_dir/../genome.config | sed "s/^\([A-Za-z0-9_]*\).*/$QPALMA_contigs,\'\1\'/g"`
done
QPALMA_contigs=`echo "$QPALMA_contigs" | sed "s/^,//g"`
export QPALMA_contigs

export QPALMA_genome_result_dir=$result_dir/genome
mkdir -p $result_dir/genome
$QPALMA_PATH/galaxy/prep_genome.sh $QPALMA_genome_dir/../genome.config $result_dir/genome

sed_QPALMA_contigs=`echo "$QPALMA_contigs"|tr "\n" "," | sed "s/,$//g"`
export sed_QPALMA_contigs

echo \#\#QPALMA parameter file \(version 0.9.3\) 
echo \#\# 
echo \#\# configuration: 
echo \#\#   QPALMA_C=$QPALMA_C
echo \#\#   QPALMA_half_window_size=$QPALMA_half_window_size 
echo \#\#   QPALMA_prb_offset=$QPALMA_prb_offset
echo \#\#   QPALMA_iterations=$QPALMA_iterations
echo \#\#   QPALMA_max_intron_length=$QPALMA_max_intron_length
echo \#\#   QPALMA_max_qual=$QPALMA_max_qual
echo \#\#   QPALMA_num_supp_points=$QPALMA_num_supp_points
echo \#\#   QPALMA_enable_intron_length=$QPALMA_enable_intron_length
echo \#\#   QPALMA_enable_splice_scores=$QPALMA_enable_splice_scores
echo \#\#   QPALMA_drop_unspliced=$QPALMA_drop_unspliced
echo \#\#   QPALMA_max_unspliced_reads=$QPALMA_max_number_unspliced
echo \#\#   QPALMA_contigs=$sed_QPALMA_contigs
echo \#\# 

if [ $QPALMA_enable_splice_scores == True ]
then
    cat $QPALMA_PATH/galaxy/galaxy.conf | sed "s/QPALMA_C/$QPALMA_C/g" | \
	sed "s/QPALMA_half_window_size/$QPALMA_half_window_size/g" | \
	sed "s/QPALMA_prb_offset/$QPALMA_prb_offset/g" | \
	sed "s/QPALMA_iterations/$QPALMA_iterations/g" | \
	sed "s/QPALMA_max_intron_length/$QPALMA_max_intron_length/g" | \
	sed "s/QPALMA_max_qual/$QPALMA_max_qual/g" | \
	sed "s/QPALMA_num_supp_points/$QPALMA_num_supp_points/g" | \
	sed "s/QPALMA_enable_intron_length/$QPALMA_enable_intron_length/g" | \
	sed "s/QPALMA_enable_splice_scores/$QPALMA_enable_splice_scores/g" | \
	sed "s/QPALMA_drop_unspliced/$QPALMA_drop_unspliced/g" | \
	sed "s/QPALMA_max_unspliced_reads/$QPALMA_max_number_unspliced/g" | \
	sed "s/QPALMA_contigs/$sed_QPALMA_contigs/g" > $result_dir/galaxy.conf
else
    cat $QPALMA_PATH/galaxy/galaxy.conf | sed "s/QPALMA_C/$QPALMA_C/g" | \
	sed "s/QPALMA_half_window_size/$QPALMA_half_window_size/g" | \
	sed "s/QPALMA_prb_offset/$QPALMA_prb_offset/g" | \
	sed "s/QPALMA_iterations/$QPALMA_iterations/g" | \
	sed "s/QPALMA_max_intron_length/$QPALMA_max_intron_length/g" | \
	sed "s/QPALMA_max_qual/$QPALMA_max_qual/g" | \
	sed "s/QPALMA_num_supp_points/$QPALMA_num_supp_points/g" | \
	sed "s/QPALMA_enable_intron_length/$QPALMA_enable_intron_length/g" | \
	sed "s/acceptor_scores_loc/#acceptor_scores_loc/g" | \
	sed "s/donor_scores_loc/#donor_scores_loc/g" | \
	sed "s/QPALMA_enable_splice_scores/$QPALMA_enable_splice_scores/g" | \
	sed "s/QPALMA_drop_unspliced/$QPALMA_drop_unspliced/g" | \
	sed "s/QPALMA_max_unspliced_reads/$QPALMA_max_number_unspliced/g" | \
	sed "s/QPALMA_contigs/$sed_QPALMA_contigs/g" > $result_dir/galaxy.conf
fi

set +e
echo "Limitation of the number of spliced and unspliced reads for training"
let spliced_reads=5000
let unspliced_reads=2500
let num_unspliced=$QPALMA_max_number_unspliced
if [ "$num_unspliced" -lt "$unspliced_reads" ]
then
    let unspliced_reads=$num_unspliced
fi

let num_spliced=`cut -f11 $data_file | grep "," | wc -l`
#echo $num_spliced spliced reads in $data_file
if [ "$num_spliced" -lt "$spliced_reads" ]
then
    let spliced_reads=$num_spliced
    if [ "$unspliced_reads" -gt "$(($spliced_reads/2))" ]
    then
	let unspliced_reads=$spliced_reads/2
    fi
fi


if [ -e $data_file.filt ]
then
    rm $data_file.filt
fi


echo "Training set composed of $spliced_reads spliced reads and $unspliced_reads unspliced reads"
while read line
do
	#echo "$line"
    if [ "`echo "$line" | cut -f11 | grep ","`" != "" ]
    then
		if [ "$spliced_reads" -gt "0" ]
		then
			echo "$line" | tr " " "\t" >> $data_file.filt
			let spliced_reads=$spliced_reads-1
		fi
    else
		if [ "$unspliced_reads" -gt "0" ]
		then
			echo "$line" | tr " " "\t" >> $data_file.filt
			let unspliced_reads=$unspliced_reads-1
		fi
    fi
done < "$data_file"

echo "done..."
set -e

##Launch training
$QPALMA_PATH/bin/qpalma_train $result_dir/galaxy.conf $data_file.filt $result_dir

echo \#\#QPALMA parameter file \(version 0.9.3\) > $3
echo \#\# >> $3
echo \#\# configuration: >> $3 
echo \#\#   QPALMA_C=$QPALMA_C >> $3
echo \#\#   QPALMA_half_window_size=$QPALMA_half_window_size >> $3
echo \#\#   QPALMA_prb_offset=$QPALMA_prb_offset>> $3
echo \#\#   QPALMA_iterations=$QPALMA_iterations>> $3
echo \#\#   QPALMA_max_intron_length=$QPALMA_max_intron_length>> $3
echo \#\#   QPALMA_max_qual=$QPALMA_max_qual>> $3
echo \#\#   QPALMA_num_supp_points=$QPALMA_num_supp_points>> $3
echo \#\#   QPALMA_enable_intron_length=$QPALMA_enable_intron_length>> $3
echo \#\#   QPALMA_enable_splice_scores=$QPALMA_enable_splice_scores>> $3
echo \#\#   QPALMA_drop_unspliced=$QPALMA_drop_unspliced>> $3
echo \#\#   QPALMA_max_unspliced_reads=$QPALMA_max_number_unspliced>> $3
echo \#\#   QPALMA_contigs=$QPALMA_contigs>> $3
echo \#\# >> $3

cat $result_dir/parameters_final.qpalma >> $3


if [ "$QPALMA_genome_fasta" != "" ]
then
    rm -rf $QPALMA_genome_dir_base
fi
