/*********************************************************************************************/
// Version of fill_matrix with fills several alignment matrices (best, second_best ...)
// nr_paths det. number of these matrices
//
// est_len: Number of columns in alignment matrix: length(EST) +1 (for the gap) (variable: i)
// dna_len: Number of lines in alignment matrix: length(DNA) +1 (for the gap) (variable: j)
//
// !!! IMPORTANT !!! The pictures and equations in the paper/poster etc. describe the algorithm
// for an alignment matrix where DNA and EST changed places !!! IMPORTANT !!!
/*********************************************************************************************/

/*
  matchmatrix: columnwise
    - A C G T N (dna)
  - x x x x x x
  A x x x x x x	
  C x x x x x x
  G x x x x x x
  T x x x x x x
  N x x x x x x
(est)
*/

/*
  alignment matrix: columnwise
   |-EST . . .
  -+------------
  -|00 ...
  D|0.
  N|. .
  A|.  .
  .|.
  .|
  .|
*/


/* TODO
   - max_len = 10000 for elegans!
   - do not use double
*/

#include "fill_matrix.h"
#include "debug_tools.h"

#define D_USE_QUALITY_SCORES 1

int check_char(char base)
{
  switch(base)
  {
  case 'A':
  case 'a': return 1;
  case 'C': 	
  case 'c': return 2;
  case 'G':
  case 'g': return 3;
  case 'T':
  case 't': return 4;
  case 'N':
  case 'n': return 5; 
  //default : return 0;
  default : return -1;
  }
  
}

 /* system of Palma. Otherwise scoring_functions should point to an array of
 * plifs scoring the respective pairs of characters together with the EST
 * quality score.
 */

inline double getScore(struct penalty_struct* qualityScores, int mlen, int dnaChar, int estChar, double baseScore) {
   double score = 0.0;
   FA(estChar > 0 && estChar < 6);
   FA(dnaChar > -1 && dnaChar < 6);

   //printf("Starting scoring / scheme is USE_QUALITY_SCORES\n");
   //printf("estChar,dnaChar are: %d,%d,%d",estChar,dnaChar);
   
   int currentPos = (estChar-1)*6+dnaChar;
   struct penalty_struct currentPen = qualityScores[currentPos];

   //if( estChar == 1 and dnaChar == 1)
   //   printf("A-A : %f %f\n",currentPen.penalties[0],currentPen.penalties[1]);

	INT idx = 0 ;
	for (INT i=0; i<currentPen.len; i++)
		if (currentPen.limits[i] <= baseScore)
			idx++ ;
	
	if (idx==0)
		score=currentPen.penalties[0] ;
	else if (idx==currentPen.len)
		score=currentPen.penalties[currentPen.len-1] ;
	else
	{
		score = (currentPen.penalties[idx]*(baseScore-currentPen.limits[idx-1]) + currentPen.penalties[idx-1]*
			   (currentPen.limits[idx]-baseScore)) / (currentPen.limits[idx]-currentPen.limits[idx-1]) ;  
   }

   //score = lookup_penalty(&currentPen,idx, NULL, 0);
   //if(baseScore <= currentPen.limits[0]) {
   //   score = currentPen.penalties[0];
   //} else {
   //   int sp_idx;
   //   for(sp_idx=0;sp_idx<currentPen.len;sp_idx++) {
   //      if(currentPen.limits[sp_idx] <= baseScore )
   //         score = currentPen.penalties[sp_idx];
   //   //printf("Current score is %f\n",score);
   //   }
   //}

   return score;
}

void fill_matrix(int nr_paths, Pre_score* matrices[], int est_len /*counter var: i*/, int dna_len /*counter var: j*/, char* est, char* dna, double* prb, penalty_struct* functions , double* matchmatrix, penalty_struct* qualityScores, double* donor, double* acceptor, bool remove_duplicate_scores, int nr_donor_sites, int* donor_sites, int* max_score_positions)
{	
  //printf("Entering fill_matrix...\n");

  /*********************************************************************************************/
  // variables
  /*********************************************************************************************/
  const int mlen = 6; // length of matchmatrix
  int estChar, dnaChar; // 0:'-', 1:'a', 2:'c', 3:'g', 4:'t', 5:'n'

  double baseScore;
  double *est_scores = prb ;

  double prevValue ;
  double tempValue;
  int max_len = (int)functions->limits[functions->len-1] ; //last (len) supporting point of h
  //printf("max_len: %d\n", max_len) ;

  int num_elem ; // counter variable

  const int max_num_elem = (4+nr_paths+max_len) * nr_paths ; // upper bound on num_elem
  double *scores=new double[max_num_elem] ; //are sorted later
  struct ArrayElem *array=new ArrayElem[max_num_elem] ;
  //printf("%d\n", max_num_elem*sizeof(ArrayElem)) ;
  
  int max_ss_element[nr_paths*nr_paths]; //position on dna with max ss score for introns of length greater than max_len
  int splice_pos ; //splice position get from donor_sites
  int start_splice_pos ; //where to start looking in donor_sites

  double max_scores[nr_paths] ;

  /*********************************************************************************************/
  // Initialisation of all alignment matrices, columnwise
  /*********************************************************************************************/
  for (int z=0; z<nr_paths; z++) /* matrice z */
  {
    max_scores[z] = log(0.0) ;
    max_score_positions[2*z]    = 0; // pos i
    max_score_positions[2*z +1] = 0; // pos j

    for (int i=0; i<est_len; i++) /* i.column */
    {
      for (int j=0; j<dna_len; j++) /* j.line */
      {
	((Pre_score*)matrices[z] +(i*dna_len) +j)->value = log(0.0); // -inf
	((Pre_score*)matrices[z] +(i*dna_len) +j)->prev_i = 0;
	((Pre_score*)matrices[z] +(i*dna_len) +j)->prev_j = 0;
	((Pre_score*)matrices[z] +(i*dna_len) +j)->prev_matrix_no = 0;
	((Pre_score*)matrices[z] +(i*dna_len) +j)->issplice = 0;
      }
    }
  }

  /*********************************************************************************************/	
  // Element (0,0) for first/best matrice (z=0)
  /*********************************************************************************************/	 
  ((Pre_score*)matrices[0])->value = 0;
  ((Pre_score*)matrices[0])->prev_i = 0;
  ((Pre_score*)matrices[0])->prev_j = 0;
  ((Pre_score*)matrices[0])->prev_matrix_no = 0;
  ((Pre_score*)matrices[0])->issplice = 0;
 

  /*********************************************************************************************/	
  // Line Zero for first/best matrice (z=0)
  /*********************************************************************************************/
  for (int i=1; i<est_len; i++) /* 0.line, all columns (DNA='-') */
  { 	
    ((Pre_score*)matrices[0] +(i*dna_len))->value = 0 ;		
    ((Pre_score*)matrices[0] +(i*dna_len))->prev_i = i-1;
    ((Pre_score*)matrices[0] +(i*dna_len))->prev_j = 0;    
    ((Pre_score*)matrices[0] +(i*dna_len))->prev_matrix_no = 0;
    ((Pre_score*)matrices[0] +(i*dna_len))->issplice = 0;
  }


  /*********************************************************************************************/	
  // Column Zero for first/best matrice (z=0)
  /*********************************************************************************************/	
  for (int j=1; j<dna_len; j++) // 0.column, all lines
  {      
    ((Pre_score*)matrices[0] +j)->value =  0 ;
    ((Pre_score*)matrices[0] +j)->prev_i = 0 ;
    ((Pre_score*)matrices[0] +j)->prev_j = j-1 ;
    ((Pre_score*)matrices[0] +j)->prev_matrix_no = 0 ;
    ((Pre_score*)matrices[0] +j)->issplice = 0;
  }

   
  /*********************************************************************************************/	
  // !!!! Column Zero and line Zero for all other matrices except the best one (z>0): -INF !!!! 
  /*********************************************************************************************/	


   //printf("after initialization\n");
  
  /*********************************************************************************************/
  // all other entries for all matrices, last column and row treated seperately
  /*********************************************************************************************/
  
  /* columnwise, 0.column and 0.line already filled */
  for (int i=1; i<est_len; i++) /* i.column */
  {
    
    start_splice_pos = 0 ; //where to start looking in donor_sites
    for (int zz=0; zz<nr_paths*nr_paths; zz++) {max_ss_element[zz]=0 ;} // initializing
  
    for (int j=1; j<dna_len; j++) //j.line
    {
      dnaChar = check_char(dna[j-1]) ; // -1 because of of the gaps in the 0.line/column of the al-matrix
      estChar = check_char(est[i-1]) ;
      baseScore = est_scores[i-1];

      //printf("at iteration %d %d\n",i,j);

      //if introns of length greater than max_len possible
      if (j > max_len)
      {
	//storing highest ss_score (the position) and NOT -INF
	int new_ss_pos = j-max_len ;
	double donor_score = donor[new_ss_pos-1] ;
	if (isfinite(donor_score)) //splice site!
        {
	  if (max_ss_element[nr_paths-1] != 0)
          { 
	    for (int zz=0; zz<nr_paths; zz++) //for each alignment matrix
            {
	      int pos_with_minscore = nr_paths ; //assuming, the new score is the worst score
	      double new_intron_score = donor_score+((Pre_score*)matrices[zz]+(i*dna_len)+new_ss_pos-1)->value ; //could be -INF
	      if (!(isfinite(new_intron_score))) {printf("new_intron_score is -INF\n") ;}
	      for (int pos=0; pos<nr_paths; pos++) //find worst score of all nr_paths+1 scores
              {
		int act_splice_position = max_ss_element[zz*nr_paths+pos] ;
		assert(isfinite(donor[act_splice_position-1])) ;
		double act_intron_score = donor[act_splice_position-1]+((Pre_score*)matrices[zz]+(i*dna_len)+act_splice_position-1)->value ;
		if (new_intron_score >= act_intron_score) 
                {
		  new_intron_score = act_intron_score ;
		  pos_with_minscore = pos ;
		} //else nothing happens, new intron score still the worst one
	      }
	      if (pos_with_minscore != nr_paths)
              {
		max_ss_element[zz*nr_paths+pos_with_minscore] = new_ss_pos ; //replacing
	      }
	    }
	  }
	  else
          {
	    for (int pos=0; pos<nr_paths; pos++)
            {
	      if (max_ss_element[pos]==0)  //no previous splice site at this pos
              {
		for (int zz=0; zz<nr_paths; zz++) {max_ss_element[zz*nr_paths+pos]=j-max_len ;}
		break ;
	      }
	    } 
	  }
	}
	//start_pos in donor_sites so that only introns shorter then max_length are seen
	if ((start_splice_pos < nr_donor_sites) && donor_sites[start_splice_pos] == j-max_len)
        {
	  start_splice_pos++ ;
	}
      }
      

      /* Computing all possible scores */
      num_elem = 0 ; /* number of different scores/possibilities so far */
      for (int z=0; z<nr_paths; z++) /* matrice z */
      {
	Pre_score* actMatrix = (Pre_score*)matrices[z]; /* hm, why?  */

	/***************************************/
	// 0. Zero (local alignment version)
        /***************************************/
	assert(num_elem<max_num_elem) ;
	array[num_elem].prev_i = 0; 
	array[num_elem].prev_j = 0; 
	array[num_elem].prev_matrix_no = z;
	array[num_elem].isSplice = false; /* no splice site */
	scores[num_elem] = 0 ;
	num_elem++ ;


        /***************************************/
	// 1. gap on EST (j-1,i) -> (j,i) 
        /***************************************/
	prevValue = ((Pre_score*)actMatrix +(i*dna_len)+j-1)->value ; /* (j-1,i) -> (j,i) */
	if (isfinite(prevValue))
	{
      //if( currentMode == USE_QUALITY_SCORES ) {
#if D_USE_QUALITY_SCORES 
         tempValue = prevValue + matchmatrix[dnaChar];
#else
      //} else {
	      tempValue = prevValue + matchmatrix[mlen*dnaChar];   /* score(gap,DNA) */
      //}
#endif

     //printf("tempValue: %f\n",tempValue);

	  if (isfinite(tempValue))
	  {
	    assert(num_elem<max_num_elem) ;
	    array[num_elem].prev_i = i; 
	    array[num_elem].prev_j = j-1; /* predecessor */
	    array[num_elem].prev_matrix_no = z;
	    array[num_elem].isSplice = false; /* no splice site */
	    scores[num_elem] = -tempValue ;
	    num_elem++ ;
	  }
	}
      

	/***************************************/
    	// 2. match/mismatch (j-1,i-1) -> (j,i)
	/***************************************/
	prevValue = ((Pre_score*)actMatrix +(i-1)*dna_len + j-1)->value ; /* (j-1,i-1) -> (j,i) */
	if (isfinite(prevValue))
	{
      //if( currentMode == USE_QUALITY_SCORES ) {
#if D_USE_QUALITY_SCORES 
        tempValue = prevValue + getScore(qualityScores,mlen,dnaChar,estChar,baseScore);
#else
      //} else {
        tempValue = prevValue +(matchmatrix[mlen* dnaChar +estChar]);
#endif
     //}

     //printf("tempValue: %f\n",tempValue);

	  if (isfinite(tempValue))
	  {
	    assert(num_elem<max_num_elem) ;
	    array[num_elem].prev_i = i-1; /* predecessor */
	    array[num_elem].prev_j = j-1; /* predecessor */
	    array[num_elem].prev_matrix_no = z;
	    array[num_elem].isSplice = false; /* no splice site */
	    scores[num_elem] = -tempValue ;
	    num_elem++ ;
	  }
	}
	

	/***************************************/
	// 3. gap on DNA (j,i-1) -> (j,i)
	/***************************************/
	prevValue = ((Pre_score*)actMatrix +(i-1)*dna_len + j)->value ;
	if (isfinite(prevValue))
	{
      //if( currentMode == USE_QUALITY_SCORES ) {
#if D_USE_QUALITY_SCORES 
        tempValue = prevValue + getScore(qualityScores,mlen,0,estChar,baseScore);
#else
//      } else {
	     tempValue = prevValue +(matchmatrix[estChar]); /* score(EST,gap) */
//      }
#endif

     //printf("tempValue: %f\n",tempValue);

	  if (isfinite(tempValue))
	  {
	    assert(num_elem<max_num_elem) ;
	    array[num_elem].prev_i = i-1; /* predecessor */
	    array[num_elem].prev_j = j; 
	    array[num_elem].prev_matrix_no = z;
	    array[num_elem].isSplice = false;  /* no splice site */
	    scores[num_elem]=-tempValue ;
	    num_elem++ ;
	  }
	}

	/***************************************/
	// 4. all possible splice sites/introns
	/***************************************/
	// acceptor[j-1]: g of ag (-1 because cpp starts counting at 0)
	// donor[splice_pos-1]: g of gt/gc (-1 because cpp starts counting at 0)
	// donor_sites: the g's of the gt/gc
	
	if (isfinite(acceptor[j-1]))
	{
	  
	  //the nr_paths many introns of length greater than max_len
	  for (int pos=0; pos<nr_paths; pos++)
          {
	    if (max_ss_element[z*nr_paths+pos]>0)
            {
	      assert((j-max_ss_element[z*nr_paths+pos]+1) > max_len) ;
	      splice_pos = max_ss_element[z*nr_paths+pos] ; //the g of the gt 
	      prevValue = ((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->value ; //one before the _gt
	      if (isfinite(prevValue) && (((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->issplice!=1))
              {
		int dist = j -(splice_pos-1); //_gt...ag: dist from g to g
		assert(dist>max_len) ;
		tempValue = lookup_penalty(functions,dist, NULL, 0) +donor[splice_pos-1] +acceptor[j-1] +prevValue;

       //printf("tempValue: %f\n",tempValue);

		if (isfinite(tempValue))
                {
		  assert(num_elem<max_num_elem) ;
		  array[num_elem].prev_i = i; 
		  array[num_elem].prev_j = splice_pos-1; // predecessor
		  array[num_elem].prev_matrix_no = z;
		  array[num_elem].isSplice = true; // splice site
		  scores[num_elem]=-tempValue ;
		  num_elem++ ;
		}
		else if (isfinite(lookup_penalty(functions,dist, NULL, 0)))
                {
		  printf("Something wrong with donor_sites(1000er) (intron_length: %d)\n", dist) ;
		  sleep(5000) ;
		}
	      }
	    }
	  }

          //all introns of length <= max_len
          for (int k=start_splice_pos; k < nr_donor_sites; k++) 
	  //for (int k=0; k < nr_donor_sites; k++) //also intron of length greater than 1000
	  {   
	    splice_pos = donor_sites[k] ;
	    if (splice_pos > j-4) { break ;}
	    prevValue = ((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->value ; //one before the _gt
	    
	    if (isfinite(prevValue) && (((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->issplice!=1))
	    {
	      int dist = j -(splice_pos-1); //_gt...ag: dist from g to g
	      assert(dist<=max_len) ;
	      tempValue = lookup_penalty(functions,dist, NULL, 0) +donor[splice_pos-1] +acceptor[j-1] +prevValue;
	      if (isfinite(tempValue))
	      {
		assert(num_elem<max_num_elem) ;
		array[num_elem].prev_i = i; 
		array[num_elem].prev_j = splice_pos-1; /* predecessor */
		array[num_elem].prev_matrix_no = z;
		array[num_elem].isSplice = true; /* splice site */
		scores[num_elem]=-tempValue ;
		num_elem++ ;
	      }
	      else if (isfinite(lookup_penalty(functions,dist, NULL, 0)))
	      {
		printf("Something wrong with donor_sites (intron_length: %d)\n", dist) ;
		sleep(5000) ;
	      }
	    }
	  }
	}

      } //end of z
      
      CMath::nmin<struct ArrayElem>(scores, array, num_elem, nr_paths);
      
      /* the nr_paths many highest scores are written in the nr_paths many matrices*/
      for (int z=0; z<nr_paths; z++) {
	if (z>=num_elem) {
	  //nothing happens, entry in matrices[z] at pos. (i,j) still -inf
	}
	else {
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->value = -scores[z] ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->prev_i = array[z].prev_i ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->prev_j = array[z].prev_j ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->issplice = array[z].isSplice ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->prev_matrix_no = array[z].prev_matrix_no;

	  if (-scores[z]>=max_scores[z]) { //storing the postion of the highest alignment score yet
	    max_scores[z] = -scores[z] ;
	    max_score_positions[2*z]    = i; // pos i
	    max_score_positions[2*z +1] = j; // pos j
	  }
	}
      } //end of z

    } //end of j
  } // end of i

  //free memory
  delete[] array;
  delete[] scores;

  //printf("Leaving fill_matrix...\n");
  return;
}
