
#include "qpalma_dp.h"
#include <cstring>

using namespace std;

/*[splice_align, est_align, weightMatch, alignmentscores, dnaest] = ...
  myalign([nr_paths], dna, est, {h}, matchmatrix, donor, acceptor, remove_duplicate_scores, ...
  print_matrix) */

Alignment::Alignment(int numQPlifs, int numq, bool use_qscores) {
      len = 0;
      limits = 0;
      penalties = 0;
      max_len = 0;
      min_len = 0;
      cache = 0;
      //enum ETransformType transform = T_LINEAR;
      id = 0;
      name = 0;
      use_svm = 0;

      // set ptrs to zero first
      splice_align         = 0;
      est_align            = 0;
      mmatrix_param        = 0;
      alignmentscores      = 0;
      qualityFeaturesAllPaths = 0;
      mlen = 6; // score matrix: length of 6 for "- A C G T N"

      numQualSuppPoints = numq;
      numPlifs = numQPlifs;
      use_quality_scores = use_qscores;

      //printf("number of support points: %d\n",numQualSuppPoints);
      //printf("number of plifs: %d\n",numPlifs );
      FA( numQualSuppPoints >= 0 );
      FA( numPlifs >= 0 );
}

void Alignment::getDNAEST(){}

void Alignment::myalign(int nr_paths_p, char* dna, int dna_len_p, char* est,
      int est_len_p, double* prb, double* chastity, struct penalty_struct h, double* matchmatrix, int mm_len,
      double* donor, int d_len, double* acceptor, int a_len, struct penalty_struct* qualityScores, 
      bool remove_duplicate_scores, bool print_matrix) {

   // printf("Entering myalign...\n");
   dna_len =  dna_len_p + 1 ;
   est_len =  est_len_p + 1 ;
   nr_paths = nr_paths_p;

  /***************************************************************************/ 
  // initialize alignment matrices and call fill_matrix()  
  /***************************************************************************/

  mode currentMode;
  if (use_quality_scores) 
   currentMode = USE_QUALITY_SCORES;
  else
   currentMode = NORMAL;

    // dnaest
   DNA_ARRAY = 0;
   EST_ARRAY = 0;

  //printf("[qpalma_dp] read is: %s\n",est);
  
  //possible donor positions
  int nr_donor_sites = 0 ;
  for (int ii=0; ii<d_len; ii++) {
    if(isfinite(donor[ii])) {
      nr_donor_sites++ ;
    }
  }

  int* donor_sites = new int[nr_donor_sites];
  int donor_idx = 0 ;
  for (int ii=0; ii<d_len; ii++) {
    if(isfinite(donor[ii])) {
      donor_sites[donor_idx] = ii+1 ;
      donor_idx++ ;
    }  
  }

  int* max_score_positions = new int[nr_paths*2];

  Pre_score** matrices = new Pre_score*[nr_paths];
  //for (int z=0; z<nr_paths; z++) {
  for (size_t  z=0; z<nr_paths; z++) {
      matrices[z] = new Pre_score[dna_len * est_len];
  }

   //printf("qualityScores\n");
   //for(int qidx=0;qidx<numPlifs;qidx++) {
   //   penalty_struct p; 
   //   p = qualityScores[qidx];
   //   printf("%d ",qidx);
   //   for(int pidx=0;pidx<p.len;pidx++)
   //      printf("%f ",pidx,p.penalties[pidx]);
   //   printf("\n");
   //}

   //int h_idx;
   //for(h_idx=0;h_idx<h.len;h_idx++)
   //   printf("%d %f %f\n",h_idx,h.limits[h_idx],h.penalties[h_idx]);
  
  //printf("calling fill_matrix...\n");
  fill_matrix(nr_paths, matrices, est_len, dna_len, est, dna, prb, &h, matchmatrix, qualityScores, donor, acceptor, remove_duplicate_scores, nr_donor_sites, donor_sites, max_score_positions);
  //printf("after call to fill_matrix...\n");

  /***************************************************************************/ 
  // return arguments etc.
  /***************************************************************************/
  int result_length; //Eine Variable fuer alle Matrizen

   splice_align_size = (dna_len-1)*nr_paths;
   est_align_size = (est_len-1)*nr_paths;

   int mmatrix_size;

   if (currentMode == USE_QUALITY_SCORES) {
      mmatrix_param_size = mlen*nr_paths;
      mmatrix_size = mlen;
   }

   if (currentMode == NORMAL) {
      mmatrix_param_size = (mlen*mlen)*nr_paths;
      mmatrix_size = mlen*mlen;
   }
   
   alignmentscores_size = nr_paths; //alignment score for each path/matrix
   numPathsPlifs = numPlifs*nr_paths; //alignment score for each path/matrix

   splice_align         = new int[splice_align_size];
   est_align            = new int[est_align_size];
   mmatrix_param        = new int[mmatrix_param_size];
   alignmentscores      = new double[nr_paths]; //alignment score for each path/matrix


  // printf("before memset...\n");
  memset((char*)splice_align, -1, (dna_len-1)*nr_paths*sizeof(int)); // fills splice_align with zeros
  memset((char*)est_align, -1, (est_len-1)*nr_paths*sizeof(int)); // fills est_align with zeros
  memset((char*)mmatrix_param, 0, mmatrix_size*nr_paths*sizeof(int)); //fills mmatrix_param with zeros
  memset(alignmentscores, -1, nr_paths*sizeof(double)); //fills alignmentscores with zeros
  //printf("after memset...\n");
  
  qualityFeaturesAllPaths= new penalty_struct*[nr_paths];

  //for (int z=0; z<nr_paths; z++) {
  for (size_t z=0; z<nr_paths; z++) {
    result_length = 0 ;
    
    int* s_align = splice_align + (dna_len-1)*z;  //pointer
    int* e_align = est_align + (est_len-1)*z ;    //pointer
    int* mparam  = mmatrix_param + mmatrix_size*z; //pointer

   qualityFeaturesAllPaths[z] = new penalty_struct[numPlifs];

   int qidx, pidx;
   for(qidx=0;qidx<numPlifs;qidx++) {
      penalty_struct p;
      init_penalty_struct(p);
      p.len = numQualSuppPoints;
      p.limits = (double*) calloc(p.len,sizeof(double));

      for(pidx=0;pidx<p.len;pidx++)
         p.limits[pidx] = qualityScores[qidx].limits[pidx];

      p.penalties = (double*) calloc(p.len,sizeof(double));
      qualityFeaturesAllPaths[z][qidx] = p;
   }

   //penalty_struct* qparam = qualityScoresAllPaths + (numPlifs*z);

    //printf("before call to result_align...\n");
    //bool no_more_path = 
	result_align(matrices, z, est_len, dna_len, &result_length, est, dna, prb, chastity, s_align, e_align, mparam, alignmentscores, max_score_positions, qualityFeaturesAllPaths[z] , currentMode);
    //printf("after call to result_align...\n");

   //printf("z is %d\n",z);
   //int len;
   //for(qidx=0;qidx<numPlifs;qidx++) {
   //   penalty_struct p;
   //   p = qualityScoresAllPaths[z][qidx];
   //   printf("%d: ",qidx);
   //   for(pidx=0;pidx<p.len;pidx++)
   //      printf("%f ",p.limits[pidx]);
   //   printf("\n");

   //   for(pidx=0;pidx<p.len;pidx++)
   //      printf("%f ",p.penalties[pidx]);
   //   printf("\n");
   //}

   if(z==0) {

   if(DNA_ARRAY != 0) {
       delete[] DNA_ARRAY;
       delete[] EST_ARRAY;
    }

   result_len = result_length;
   
   DNA_ARRAY = new int[result_length];
   EST_ARRAY = new int[result_length];

    //backtracking
    int i = max_score_positions[2*z] ; //i (est)
    int j = max_score_positions[2*z +1] ; //j (dna)
    int prev_i = ((Pre_score*)matrices[z] + i*dna_len +j)->prev_i; 
    int prev_j = ((Pre_score*)matrices[z] + i*dna_len +j)->prev_j;
    int prev_m_no = ((Pre_score*)matrices[z] + i*dna_len +j)->prev_matrix_no;
    
    for (int ii=result_length; ii>0; ii--) {
      if ((prev_i == (i-1)) && (prev_j == (j-1))) { // match or mismatch
	DNA_ARRAY[ii-1] = check_char(dna[j-1]) ;
	EST_ARRAY[ii-1] = check_char(est[i-1]) ;
      }
      else if ((prev_i == (i)) && (prev_j == (j-1))) {// gap on EST_ARRAY
	DNA_ARRAY[ii-1] = check_char(dna[j-1]) ;
	EST_ARRAY[ii-1] = 0 ; //gap
      }
      else if ((prev_i == (i-1)) && (prev_j == (j))) { // gap on DNA_ARRAY
	DNA_ARRAY[ii-1] = 0 ; //gap
	EST_ARRAY[ii-1] = check_char(est[i-1]) ;
      }
      else {// splice site
	for (; j > prev_j; j--) {
	  DNA_ARRAY[ii-1] = check_char(dna[j-1]) ;
	  EST_ARRAY[ii-1] = 6 ; //intron
	  ii-- ;
	}
	ii++ ; // last ii-- too much (because done in for-loop)	
      }
      
      i = prev_i;
      j = prev_j;    
      prev_i = ((Pre_score*)matrices[prev_m_no] + i*dna_len + j)->prev_i; 
      prev_j = ((Pre_score*)matrices[prev_m_no] + i*dna_len + j)->prev_j; 
      prev_m_no = ((Pre_score*)matrices[prev_m_no] + i*dna_len + j)->prev_matrix_no;
    }

    } // end of "if z == 0"
    
   } //end of z

   for (int i=nr_paths-1;i>=0; i--)
      delete[] matrices[i];
}

void Alignment::getAlignmentResults(int* s_align, int* e_align,
      int* mmatrix_p, double* alignscores, double* qScores) {

   size_t idx;
   for(idx=0; idx<splice_align_size; idx++)
      s_align[idx] = splice_align[idx];

   for(idx=0; idx<est_align_size; idx++)
      e_align[idx] =  est_align[idx];

   for(idx=0; idx<mmatrix_param_size; idx++)
      mmatrix_p[idx] = mmatrix_param[idx];

   for(idx=0; idx<alignmentscores_size; idx++)
      alignscores[idx] = alignmentscores[idx];
   
   if (use_quality_scores) {
      penalty_struct currentPlif;
      int ctr=0;
      //for (int z=0; z<nr_paths; z++) {
      for (size_t z=0; z<nr_paths; z++) {
         for(int estChar=1;estChar<6;estChar++) {
            for(int dnaChar=0;dnaChar<6;dnaChar++) {

               int currentPos = (estChar-1)*6+dnaChar;
               currentPlif = qualityFeaturesAllPaths[z][currentPos];

               for(int pidx=0; pidx<currentPlif.len; pidx++) {
                  qScores[ctr] = currentPlif.penalties[pidx];
                  //printf("%f ",qScores[ctr]);
                  ctr++;
               }
               //printf("\n");
      }}}

      //printf("\nctr is %d\n",ctr);
   }
   //printf("Leaving getAlignmentResults...\n");
}

void Alignment::getAlignmentArrays(int* dna_align, int* est_align) {

   size_t idx;
   for(idx=0; idx<result_len; idx++) {
      dna_align[idx] = DNA_ARRAY[idx];
      est_align[idx] = EST_ARRAY[idx];
   }
}


PyObject* Alignment::getAlignmentResultsNew() {

   PyObject* py_splice_align  = PyList_New(splice_align_size);
   PyObject* py_est_align     = PyList_New(est_align_size);
   PyObject* py_mmatrix       = PyList_New(mmatrix_param_size);
   PyObject* py_align_scores  = PyList_New(alignmentscores_size);
   PyObject* py_q_scores      = PyList_New(0);

   size_t idx;
   for(idx=0; idx<splice_align_size; idx++)
      PyList_SetItem(py_splice_align,idx,PyInt_FromLong(splice_align[idx]));
   //s_align[idx] = splice_align[idx];

   for(idx=0; idx<est_align_size; idx++)
      PyList_SetItem(py_est_align,idx,PyInt_FromLong(est_align[idx]));

   //e_align[idx] =  est_align[idx];

   for(idx=0; idx<mmatrix_param_size; idx++)
      PyList_SetItem(py_mmatrix,idx,PyInt_FromLong(mmatrix_param[idx]));

   //mmatrix_p[idx] = mmatrix_param[idx];

   for(idx=0; idx<alignmentscores_size; idx++)
      PyList_SetItem(py_align_scores,idx,PyFloat_FromDouble(alignmentscores[idx]));

   //alignscores[idx] = alignmentscores[idx];
   
   if (use_quality_scores) {
      penalty_struct currentPlif;
      int ctr=0;
      for (size_t z=0; z<nr_paths; z++) {
         for(int estChar=1;estChar<6;estChar++) {
            for(int dnaChar=0;dnaChar<6;dnaChar++) {

               int currentPos = (estChar-1)*6+dnaChar;
               currentPlif = qualityFeaturesAllPaths[z][currentPos];

               for(int pidx=0; pidx<currentPlif.len; pidx++) {
                  //qScores[ctr] = currentPlif.penalties[pidx];
                  //printf("%f ",qScores[ctr]);
                  PyList_Append(py_q_scores,PyFloat_FromDouble(currentPlif.penalties[pidx]));
                  ctr++;
               }
               //printf("\n");
      }}}
      //printf("\nctr is %d\n",ctr);
   }

   PyObject* result = PyTuple_Pack(5,
   py_splice_align, py_est_align, py_mmatrix,
   py_align_scores, py_q_scores);

   return result;
}

PyObject* Alignment::getAlignmentArraysNew() {

   PyObject* py_dna_align = PyList_New(result_len);
   PyObject* py_est_align = PyList_New(result_len);

   for(size_t idx=0; idx<result_len; idx++) {
      PyList_SetItem(py_dna_align,idx,PyInt_FromLong(DNA_ARRAY[idx]));
      PyList_SetItem(py_est_align,idx,PyInt_FromLong(EST_ARRAY[idx]));
   }

   PyObject* result = PyTuple_Pack(2, py_dna_align, py_est_align);
   return result;
}
