/*

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Written (W) 2007-2009 Fabio De Bona, Andre Noll, Gunnar Raetsch
Copyright (C) 2007-2009 Max-Planck-Society

*/


#include <cstdlib>
#include <cstdio>
using namespace std;

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <time.h>
#include <assert.h>

#include "IntervalQuery.h"

#define MAXLINE 255
/* #define DEBUG 1 */

/*
void die(const char* fmt,...)
{
	char string[10 * MAXLINE];
	va_list argp;
	va_start(argp, fmt);
	sprintf(string, fmt, argp);
	printf("%s",string);
	va_end(argp);
	exit(EXIT_FAILURE);
}
*/

/*enum {E_DUMMY, E_SYNTAX};*/
enum {E_SYNTAX = 1, E_BAD_INTERVAL};

/* returns 0 if an empty interval was found else the interval size, 
   lindex and rindex are only set if return value is positive */ 
int IntervalQuery::find_interval(unsigned *pos_map, off_t num_entries,
		   unsigned begin, unsigned end, unsigned *lindex, unsigned *rindex)
{
  //printf("find_interval: %d %d\n",begin,end);

	unsigned val, mid, left = 0, right = num_entries - 1;
	assert(begin <= end);
	assert(num_entries > 0);
	if (pos_map[right] < begin || pos_map[left] > end) {
#ifdef DEBUG	
	  printf("ret 0\n\n");
#endif
          return 0; /* pos_map cannot possibly contain any values between begin and end */
	}

	for (;;) {
	  mid = (right + left) / 2; /*round down */
	  val = pos_map[mid];
#ifdef DEBUG	
	  printf("left %u, right %u, mid %u, val %u\n", left, right, mid, val);
#endif
	  if (val >= begin) {
	    if (right == mid)
	      break;
	    right = mid;
	  } else {
	    if (left == mid)
	        break;
	    left = mid;
	  }
	}
	*lindex = left;
	if (pos_map[left] < begin)
	  (*lindex)++;
#ifdef DEBUG	
	printf("left %u, right %u, mid %u, val %u\n", left, right, mid, val);
	printf("pos_map[left] %u, pos_map[left+1] %u, pos_map[left+2] %u\n",
	       pos_map[left], pos_map[left+1], pos_map[left+2]);
	printf("lindex %u, pos_map[lindex] %u\n", *lindex, pos_map[*lindex]);
#endif
	left = 0;
	right = num_entries - 1;
	for (;;) {
  	    mid = (right + left + 1) / 2; /*round up */
	    val = pos_map[mid];
#ifdef DEBUG	
	  printf("left %u, right %u, mid %u, val %u\n", left, right, mid, val);
#endif
	    if (val <= end) {
	        if (left == mid)
		    break;
		left = mid;
	    } else {
	      if (right == mid)
		  break;
	      right = mid;
	    }
	}
	*rindex = mid;
	if (pos_map[right] > end)
	  (*rindex)--;

#ifdef DEBUG	
	printf("left %u, right %u, mid %u, val %u\n", left, right, mid, val);
	printf("pos_map[right] %u, pos_map[right-1] %u, pos_map[right-2] %u\n", 
	       pos_map[right], pos_map[right-1], pos_map[right-2]);
	printf("rindex %u, pos_map[rindex] %u\n", *rindex, pos_map[*rindex]);
	printf("ret %i\n\n", *rindex - *lindex + 1);
#endif
        //assert(*rindex >= *lindex);
        return *rindex - *lindex + 1;
}


int IntervalQuery::mmap_file(const char *filename, int open_mode, void **map, off_t *size)
{
	int fd, ret, mmap_prot, mmap_flags;
	struct stat file_status;

	if (open_mode == O_RDONLY) {
		mmap_prot = PROT_READ;
		mmap_flags = MAP_PRIVATE;
	} else {
		mmap_prot = PROT_READ | PROT_WRITE;
		mmap_flags = MAP_SHARED;
	}
	/*printf("filename: %s\n", filename);*/
	ret = open(filename, open_mode);
	if (ret < 0)
	{
		fprintf(stderr, "can not open %s: %s\n", filename, strerror(errno));
		exit(EXIT_FAILURE);
	} ;

	fd = ret;
	if (fstat(fd, &file_status) < 0)
	{
		fprintf(stderr, "fstat error: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	*size = file_status.st_size;
	if (!*size)
	{
		fprintf(stderr, "mmap error: %s is empty\n", filename);
		exit(EXIT_FAILURE);
	}

	*map = mmap(NULL, *size, mmap_prot, mmap_flags, fd, 0);
	if (*map == MAP_FAILED)
	{
		fprintf(stderr, "mmap error: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}		
	ret = 1;
	close(fd);
	return ret;
}


/*
 * rhs[0] basename
 * rhs[1] list of m score names
 * rsh[2] Interval Matrix (n x 2)
 *
 * return:
 * lhs[0]: list of positions that lie in any given interval, k := number of such positions
 * lhs[1]: k x m score Matrix
 * lhs[2]: index vector (k x 1), lhs[2][j] = l <=> rhs[2][l, 0] <= lhs[0][j] <= rhs[2][l, 1].
 */

void IntervalQuery::interval_query(char* basename, char** score_names, int num_scores, int* interval_matrix, int num_intervals)
{
   //printf("Entering interval_query...\n");
	char filename[MAXLINE], *err_txt;
	int ret;
	unsigned i, j = 0;
	float **score_maps = NULL;
	off_t *score_sizes = NULL;
	//double *interval_ptr;
	int* interval_ptr;
	unsigned m, k,total_num_pos, *pos_map = NULL;
	off_t pos_size, num_positions;

   //printf("initialized local vars...\n");

	ret = -E_SYNTAX;
	err_txt = (char*)"fatal: no interval(s) given";
	if (!num_intervals)
		goto out;
	err_txt = (char*)"fatal: no score file(s) given";
	if (!num_scores)
		goto out;

	score_maps = (float**) malloc(num_scores * sizeof(void *));
	memset(score_maps, 0, num_scores * sizeof(void *));
	score_sizes = (off_t*) malloc(num_scores * sizeof(off_t *));

   //printf("setting score_maps...\n");

	//printf("%s.pos\n", basename);
	sprintf(filename, "%s.pos", basename);
	ret = mmap_file(filename, O_RDONLY, (void **)&pos_map, &pos_size);
	err_txt = (char*)"pos mmap error";
	if (ret < 0)
		goto out;
	num_positions =	pos_size / sizeof(unsigned);
	
   //printf("Loading file / mmapping...\n");

//#ifdef DEBUG
	err_txt = (char*)"pos map is not sorted";
	//printf("num_positions: %i\n", num_positions);
	for (i=1; i<num_positions; ++i) {
	    if (pos_map[i-1] > pos_map[i])
	            goto out;
	}
	//printf("successfully checked that pos_map is sorted\n");
//#endif
	
	for (i = 0; i < num_scores; i++) {
		//printf("%s.%s\n", basename, score_names[i]);
		sprintf(filename, "%s.%s", basename, score_names[i]);
		ret = mmap_file(filename, O_RDONLY, (void **)&score_maps[i], &score_sizes[i]);
		err_txt = (char*)"score mmap error";
		if (ret < 0)
		        goto out;
	}
	total_num_pos = 0;

   interval_ptr = interval_matrix;
	for (j = 0; j < num_intervals; j++) {
	    unsigned left = (unsigned)interval_ptr[2 * j];
	    unsigned right = (unsigned)interval_ptr[2 * j + 1];
       //printf("left/right are %d/%d\n",left,right);
   
#ifdef DEBUG

	      for (i=0; i<num_positions; ++i) {
		if (left <= pos_map[i] && pos_map[i] <= right)
		  printf("%i ", pos_map[i]);
	      }
	      printf("\n\n");
#endif

	    unsigned lindex, rindex;
	    
	    ret = -E_SYNTAX;
	    err_txt = (char*)"bad interval";
	    if (left > right)
	        goto out;
	    ret = find_interval(pos_map, num_positions, left, right, &lindex, &rindex);
	    if (ret > 0) {/* non-empty interval */
	        total_num_pos += rindex - lindex + 1;
#ifdef DEBUG
		printf("lidx: %u, ridx: %u, t: %u\n", lindex, rindex, total_num_pos);
#endif
	    }
	}

   //printf("after 1st call to find_interval\n");

   //pos_ptr = (int*) malloc(total_num_pos*sizeof(int));
	//score_ptr = (double*) malloc(total_num_pos*num_scores*sizeof(double));
   //index_ptr = (int*) malloc(total_num_pos*sizeof(int));

   pos_ptr = new int[total_num_pos];
	score_ptr = new double[total_num_pos*num_scores];
   index_ptr = new int[total_num_pos];

   num_hits_p = total_num_pos;

	k = 0;
	for (j = 0; j < num_intervals; j++) {
	    unsigned left = (unsigned)interval_ptr[2 * j];
	    unsigned right = (unsigned)interval_ptr[2 * j + 1];
	    unsigned lindex, rindex;
	    
	    ret = find_interval(pos_map, num_positions, left, right, &lindex, &rindex);
	    if (ret <= 0) /* empty interval */
		continue;
            assert(ret == rindex - lindex + 1);
	    for (m = 0; m < ret; m++) {
	      if (index_ptr)
	  	index_ptr[k + m] = j;
	      err_txt = (char*)"returned position out of interval";
              if (left > pos_map[lindex + m] || pos_map[lindex + m] > right)
		goto out;

#ifdef DEBUG
              printf("l: %u, r: %u, p: %u\n", left, right, pos_map[lindex + m]);
#endif
	      pos_ptr[k + m] = pos_map[lindex + m];
	      for (i = 0; i < num_scores; i++)
		/*score_ptr[i + num_scores * (k + m)] = (double)score_maps[i][lindex + m];*/
		score_ptr[i * total_num_pos + (k + m)] = (double)score_maps[i][lindex + m];
	    }
	    k += ret;
	}

	ret = 1;

   //printf("after 2nd call to find_interval\n");

 out:
	for (i = 0; i < num_scores; i++) {
	  if (score_maps[i])
	    munmap(score_maps[i], score_sizes[i]);
	}

	free(score_maps);
	free(score_sizes);
	if (pos_map)
	  munmap(pos_map, pos_size);

   //printf("Leaving interval_query...\n");
}

int IntervalQuery::query(char* basename, char** score_names, int num_scores, int* interval_matrix, int num_intervals) {

         interval_query(basename, score_names, num_scores, interval_matrix, num_intervals);

         num_scores_p = num_scores;

         return num_hits_p;
}


void IntervalQuery::getResults(int* pos, int* index, double* score) {
   
   for(int k=0; k<num_hits_p;k++) {
      pos[k] = pos_ptr[k];
      index[k] = index_ptr[k];
   }
      
   for(int k=0; k<num_hits_p*num_scores_p;k++) {
      score[k] = score_ptr[k];
   }

}
