#include "ParaParser.h"
#include <stdio.h>
#include <string>
using namespace std;


#ifdef __APPLE__
#include "getline.h"
#endif

/*
 *
 *
 */

/*
bool check_for_well_formed_format_string(const char* fmt) {
   types_list = (char**) malloc(sizeof(char**)*num_columns);
   char *pruned_ptr = pruned_format_string;
   //printf("types list\n");
   for(size_t f_idx=0;f_idx<num_columns;f_idx++) {
      char *part = strtok (pruned_ptr, "%");
      pruned_ptr = NULL;

}
*/

/**
 * Split string and return pointers to its parts.
 *
 * \param args The string to be split.
 * \param argv_ptr Pointer to the list of substrings.
 * \param delim Delimiter.
 *
 * This function modifies \a args by replacing each occurance of \a delim by
 * zero. A \p NULL-terminated array of pointers to char* is allocated dynamically
 * and these pointers are initialized to point to the broken-up substrings
 * within \a args. A pointer to this array is returned via \a argv_ptr.
 *
 * \return The number of substrings found in \a args.
 */

unsigned split_args(char *args, char *** const argv_ptr, const char *delim)
{
   char *p = args;
   char **argv;
   size_t n = 0, i, j;

   p = args + strspn(args, delim);
   for (;;) {
      i = strcspn(p, delim);
      if (!i)
         break;
      p += i;
      n++;
      p += strspn(p, delim);
   }
   *argv_ptr = (char**) malloc((n + 1) * sizeof(char *));
   argv = *argv_ptr;
   i = 0;
   p = args + strspn(args, delim);
   while (p) {
      argv[i] = p;
      j = strcspn(p, delim);
      if (!j)
         break;
      p += strcspn(p, delim);
      if (*p) {
         *p = '\0';
         p++;
         p += strspn(p, delim);
      }
      i++;
   }
   argv[n] = NULL;
   return n;
}


/*
 * The constructor needs the format string to be used with sscanf and the names
 * of the respective fields for the dictionary.
 *
 */

ParaParser::ParaParser(const char* fmt, char** _fields, int num_entries, storage_mode mode) {
   // check that we have more than zero entries and that the format string
   // contains exactly num_entries format elements.
   if ( num_entries < 1 )
      printf("Error: You need at least one field !\n");
      
   num_columns = (size_t) num_entries;

   // count how many entries are parsed in one line (number of %'s)
   size_t format_num_entries = 0;
   for(size_t fidx=0;fidx<strlen(fmt);fidx++)
      if (fmt[fidx] == '%')
         format_num_entries++;

   if (format_num_entries != num_columns) {
      printf("Error: For every entry in the format string you have to supply a name!\n");
      exit(EXIT_FAILURE);
   }

   // copy the field names to the member variable
   size_t buf_size = 512;
   field_names = (char**) malloc(sizeof(char*)*num_columns);
   for(size_t idx=0;idx<num_columns;idx++) {
      field_names[idx] = (char*) malloc(sizeof(char)*buf_size);
      strncpy(field_names[idx],_fields[idx],strlen(_fields[idx])+1);
      field_names[idx][strlen(_fields[idx])] = '\0';
      //printf("%s\n",field_names[idx]);
   }

   char* pruned_format_string = (char*) malloc(sizeof(char)*buf_size);
   size_t pruned_size = 0;
   for(size_t idx=0;idx<strlen(fmt);idx++) {
      pruned_format_string[pruned_size] = fmt[idx];
      pruned_size++;
   }

   pruned_format_string[pruned_size] = '%';
   pruned_format_string[++pruned_size] = '\0';
   //printf("%s\n",pruned_format_string);

   types_list = (char**) malloc(sizeof(char**)*num_columns);
   char *pruned_ptr = pruned_format_string;
   //printf("types list\n");
   for(size_t f_idx=0;f_idx<num_columns;f_idx++) {
      char *part = strtok (pruned_ptr, "%");
      pruned_ptr = NULL;

      types_list[f_idx] = (char*) malloc(sizeof(char*)*strlen(part)+1);
      types_list[f_idx][strlen(part)] = '\0';
      strncpy(types_list[f_idx],part,strlen(part));
      //printf("%s(%d) ",part,strlen(part));
   }
   //printf("\n");
   //
   current_mode = mode;
   if(current_mode !=  IN_VECTOR && current_mode != IN_MAP) {
      printf("Error: Wrong save mode!");
      exit(EXIT_FAILURE);
   }
}



ParaParser::~ParaParser() {
   // unmap parsed file
   int status = munmap(reads_area,reads_filesize);
   if(status != 0)
      perror("munmap");
}


/*
 * 
 *
 *
 */

void ParaParser::create_entry_from_line(const char* current_line, char* _format_string, char* lineBeginPtr, char* lineEndPtr) {
   //printf("current line is %s",current_line);
   
   // create an array of void ptrs
   char **current_entries = (char**) malloc(sizeof(char*)*num_columns);

   char* mutable_line = (char*) malloc(sizeof(char)*strlen(current_line));
   strncpy(mutable_line,current_line,strlen(current_line));
   mutable_line[strlen(current_line)-1] = '\t';

   char** line_parts = NULL ;
   //size_t num_parts = split_args(mutable_line, &line_parts, "\t");

   for(size_t idx=0; idx<num_columns;idx++) {
      char* col = line_parts[idx];

      //printf("elem:%s\n",col);
      current_entries[idx] = (char*) malloc(sizeof(char)*strlen(col)+1);
      //current_entries2[idx] = string(col);
      strncpy(current_entries[idx],col,strlen(col));
      current_entries[idx][strlen(col)] = '\0';
   }
   free(mutable_line);

   map_key_t id = strtoul(current_entries[0],NULL,10);

   if ( current_mode == IN_VECTOR ) {
      //printf("size is %zd\n",v_entries->size());
      v_entries->push_back(current_entries);
      pair<char*,char*> p;
      p.first = lineBeginPtr;
      p.second = lineEndPtr;
      v_ptr_entries->push_back(p);
      //printf("size is %zd\n",v_entries->size());
   }

   if ( current_mode == IN_MAP ) {
      //printf("size is %zd\n",entries->size());
      (*entries)[id] = current_entries;
      //printf("size is %zd\n",entries->size());
   }
}


/*
 *
 *
 *
 */

int ParaParser::parseFile(char* reads_filename) {
   size_t buf_size = 512;
   char* line = (char*) malloc(sizeof(char)*buf_size);

   //printf("open %s\n",reads_filename);
   FILE *input_fs = fopen(reads_filename,"r");
   if (input_fs == NULL)
      perror("fopen");

   int line_ctr = 0;

   while (getline (&line, &buf_size, input_fs) >= 0)
      line_ctr++;

   free(line);

   //printf("file has %d lines\n",line_ctr);

   if(input_fs == NULL) {
      printf("Error: Could not open file: %s",reads_filename);
      exit(EXIT_FAILURE);
   }

   int reads_fid = fileno(input_fs);
   struct stat reads_stat;
   if ( fstat(reads_fid,&reads_stat) == -1) {
      perror("fstat");
      exit(EXIT_FAILURE);
   }

   reads_filesize = reads_stat.st_size;
   printf("Reads file is of size %lu bytes\n",(unsigned long) reads_filesize);

   if ( current_mode == IN_VECTOR ) {
      v_entries = new VECTOR();
      v_ptr_entries = new PTR_VECTOR();
   }

   if ( current_mode == IN_MAP )
      entries = new MAP();

   // try to acquire file using mmap
   reads_area = mmap (NULL,reads_filesize,PROT_READ,MAP_PRIVATE,reads_fid,0);
   if (reads_area == MAP_FAILED) {
      perror("mmap");
      exit(EXIT_FAILURE);
   }

   close(reads_fid);
   printf("Successfully mapped %lu bytes of reads file into memory\n",(unsigned long)reads_filesize);
                                      
   char* lineBeginPtr = (char*) reads_area;
   char* lineEndPtr = (char*) reads_area;     
   char* end_of_mapped_area = ((char*) reads_area) + reads_filesize;

   while (*lineEndPtr != '\n' && lineEndPtr != end_of_mapped_area) lineEndPtr++;
   lineEndPtr++;
         
   char* current_line = (char*) malloc(sizeof(char)*512);
   memset(current_line,0,512);

   unsigned long line_size = lineEndPtr - lineBeginPtr;
   strncpy(current_line,lineBeginPtr,line_size);                                                                                                                                                                               
   current_line[line_size] = '\0';

   int readCtr = 0;
   int status = 0;

   printf("Starting to parse file...\n");

   while(1) {
      if (strcmp(current_line,"") == 0) 
         break;

      create_entry_from_line(current_line,format_string,lineBeginPtr,lineEndPtr);

      if (status != 0 )
         printf("Error while parsing line (status=%d).",status);

      lineBeginPtr = lineEndPtr;                                                                                                                                                                                         
      while (*(char*)lineEndPtr != '\n' && lineEndPtr != end_of_mapped_area) lineEndPtr++;
      lineEndPtr++;

      current_line = strncpy(current_line,lineBeginPtr,lineEndPtr-lineBeginPtr);
      current_line[lineEndPtr-lineBeginPtr] = '\0';

      readCtr += 1;
   }

   printf("Successfully parsed file !\n");


   // free unneeded variables
   free(current_line);

   return readCtr;
}


/*
 *
 *
 *
 */

PyObject* ParaParser::fetchEntry(map_key_t id) {
   PyObject* line_dict = PyDict_New();

   //printf("begin of fetchEntry\n");
   //printf("size of map %d\n",entries->size());
   //printf("keys:\n");
   //MAP::iterator iter;   
   //for(iter = entries->begin(); iter != entries->end(); iter++)
   //   printf("%d\n", iter->first);
   //printf("query key is %lu\n",id);

   char** current_entry=NULL;

   char* lineBeginPtr = 0;
   char* lineEndPtr = 0;

   if ( current_mode == IN_VECTOR ) {
      //printf("IN_VECTOR mode\n");
      if (id >= v_entries->size())
         return line_dict;

      //printf("size %d\n",v_entries->size());
      current_entry = (*v_entries)[id];
      pair<char*,char*> ptr_pair = (*v_ptr_entries)[id];
      lineBeginPtr = ptr_pair.first;
      lineEndPtr = ptr_pair.second;
   }

   assert (lineBeginPtr != 0);
   assert (lineEndPtr != 0);

   if ( current_mode == IN_MAP ) {
      //printf("IN_MAP mode\n");
      MAP::iterator find_it = entries->find(id);
      if( find_it == entries->end() )
         return line_dict;

      current_entry = (*entries)[id];
   }


   int status;
      
   for(size_t idx=0;idx<num_columns;idx++) {
      char* current_type = types_list[idx];
      char* current = current_entry[idx];

      // init elem to make compiler happy
      PyObject* elem = 0;

      if ( strcmp(current_type,"d")==0 )
         elem = PyInt_FromString(current,NULL,10);

      if ( strcmp(current_type,"f")==0 )
         elem = PyFloat_FromString(PyString_FromString(current),NULL);

      if ( strcmp(current_type,"s")==0 )
         elem = PyString_FromString(current);

      if ( strcmp(current_type,"lu")==0 )
         elem = PyString_FromString(current);
         //elem = PyInt_FromString(current,NULL,10);
         //elem = PyLong_FromString(current,NULL,10);
         
      if (elem == 0)
         printf("Error: type %s/ elem %s\n",current_type,current);
      
      free(current);

      status = PyDict_SetItem(line_dict, PyString_FromString(field_names[idx]), elem);
   }
   
   /*
   //for(size_t idx=0;idx<current_read->size;idx++) {
   //   status = PyList_SetItem( prb_list, idx, PyInt_FromLong(current_read->prb[idx]-50) );
   //   status = PyList_SetItem( cal_prb_list, idx, PyInt_FromLong(current_read->cal_prb[idx]-64) );
   //   status = PyList_SetItem( chastity_list, idx, PyInt_FromLong(current_read->chastity[idx]+10) );
   //}
   */

   PyObject *return_value = PyTuple_New(2);
   PyTuple_SetItem(return_value,0,line_dict);

   char* current_line = (char*) malloc(sizeof(char)*512);
   memset(current_line,0,512);
   unsigned long line_size = lineEndPtr - lineBeginPtr;
   strncpy(current_line,lineBeginPtr,line_size-1);


   PyObject *original_line = PyString_FromString(current_line);
   PyTuple_SetItem(return_value,1,original_line);

   return return_value;
}
