# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import array
import cPickle
import numpy
import os
import os.path
import pdb
import sys

from sequence_utils import SeqSpliceInfo,DataAccessWrapper,get_flat_file_size,unbracket_seq,create_bracket_seq,reconstruct_dna_seq

jp = os.path.join

illumina_ga_range = (-5,40)
#roche_454_range   =


def processQuality(raw_qualities,prb_offset,quality_interval,perform_checks):
   """
   In order to save some space we use a signed char to store the
   qualities. Each quality element can range as follows: -128 <= elem <= 127
   """

   q_values = map(lambda x: ord(x)-prb_offset,raw_qualities)

   if perform_checks:
      for entry in q_values:
         assert quality_interval[0] <= entry <= quality_interval[1], 'Error: Your read has invalid quality values: %d' % entry

   return array.array('b',q_values)


def checkExons(dna,exons,readAlignment,exampleKey):
   """
    
   """

   fetched_dna_subseq = dna[exons[0,0]:exons[0,1]] + dna[exons[1,0]:exons[1,1]]
  
   donor_elem = dna[exons[0,1]:exons[0,1]+2]
   acceptor_elem = dna[exons[1,0]-2:exons[1,0]]

   if not ( donor_elem == 'gt' or donor_elem == 'gc' ):
      print 'checkExons: invalid donor in example %s'% exampleKey
      return False

   if not ( acceptor_elem == 'ag' ):
      print 'checkExons: invalid acceptor in example %s (%s)'% (exampleKey,dna[exons[1,0]-3:exons[1,0]+1])
      return False

   read = unbracket_seq(readAlignment)
   read = read.replace('-','')

   assert len(fetched_dna_subseq) == len(read), pdb.set_trace()

   return True

def generateTrainingDataset(in_fn,result_dir,settings,seqInfo):

   instance_counter = 0

   accessWrapper = DataAccessWrapper(settings)
   seqInfo = SeqSpliceInfo(accessWrapper,settings['allowed_fragments'])

   dataset = parse_bed_format(in_fn,settings,seqInfo,True)

   #dna = seqInfo.get_seq_and_scores(chromo,strand, seqBeginning, seqEnd, True, settings['perform_checks'])
   #relative_exons = exons - seqBeginning
   #assert checkExons(dna, relative_exons, readAlignment, id)
   #currentSeqInfo = (id,chromo,strand,seqBeginning,seqEnd)

   training_pickle_fn = jp(result_dir,'training_set.pickle')

   # saving new dataset and single keys as well
   fd=open(training_pickle_fn, 'w+')
   cPickle.dump(dataset,fd,protocol=2)
   fd.close()

   return  training_pickle_fn 


def parse_bed_format(in_fn, settings, seqInfo, train_mode=False, heuristic_mode=False):
   """
      
   """

   # This tuple specifies an interval for valid Illumina Genome Analyzer quality values
   if settings['platform'] == 'IGA':
      quality_interval = illumina_ga_range

   dataset = {}

   for line in open(in_fn):
      sl = line.strip().split('\t')
      #chromo = int(sl[0])
      chromo = sl[0]

      # for the time being we only consider chromosomes 1-5
      if not chromo in settings['allowed_fragments']:
         continue

      #pdb.set_trace()
	  
      chromStart,chromEnd =  int(sl[1]),int(sl[2])
		  
      #id       = int(sl[3])
      id       = sl[3]
      score    = sl[4]
      strand   = sl[5]

      if chromEnd<=chromStart:
          print "ignoring invalid read %s" % id
          continue 
      if chromEnd-chromStart>settings['half_window_size']:
          print "ignoring read %s spanning region of %i greater than half_window_size (%i)" % (id, chromEnd-chromStart, settings['half_window_size'])
          continue
 
      if strand in ['-','+']:
         pass
      elif strand in ['D','P']:
         strand = ['-','+'][strand == 'D']
      else:
         print 'Error strand information has to be either +/- or D/P'
         sys.exit(1)

      thickStart,thickEnd = int(sl[6]),int(sl[7])
      itemRgb     = sl[8]
      blockCount  = int(sl[9])
      blockSizes  = sl[10]
      blockStarts = sl[11]

      if settings['drop_unspliced'] and blockCount==1:
          print "dropping unspliced read %s" % id
          continue 

###      chromo = seqInfo.chromo_map[id]
###      total_size = seqInfo.chromo_sizes[chromo]
      total_size = seqInfo.chromo_sizes[seqInfo.chromo_map[chromo]]

      if train_mode:
         if blockCount > 1:
            exons = numpy.mat([0]*blockCount*2).reshape((blockCount,2))

            blockSizes = map(lambda x: int(x),blockSizes.split(','))
            blockStarts = map(lambda x: int(x),blockStarts.split(','))
          
            for idx in range(blockCount):
               exons[idx,0]     = chromStart + blockStarts[idx]
               assert(exons[idx,0]>=0), pdb.set_trace()
               assert(exons[idx,0]<=total_size), pdb.set_trace()
               exons[idx,1]     = chromStart + blockStarts[idx] + blockSizes[idx]
               assert(exons[idx,1]>=0), pdb.set_trace()
               assert(exons[idx,1]<=total_size), pdb.set_trace()
         else:
			 exons = numpy.mat([0,0]).reshape((blockCount,2))
			 blockSizes = [int(blockSizes)]
			 blockStarts = [int(blockStarts)] 
          
			 exons[0,0]     = chromStart + blockStarts[0]
			 idx = 0 ###
                         assert(exons[idx,0]>=0), pdb.set_trace()
                         assert(exons[idx,0]<=total_size), pdb.set_trace()
			 exons[0,1]     = chromStart + blockStarts[0] + blockSizes[0]
                         assert(exons[idx,1]>=0), pdb.set_trace()
                         assert(exons[idx,1]<=total_size), pdb.set_trace()
			 
      read        = sl[12]
      readQualities = processQuality(sl[13], settings['prb_offset'], quality_interval, settings['perform_checks'])

      currentSeqInfo = id,chromo,strand,chromStart,chromEnd

	  
      if train_mode:
         dataset[id] = (currentSeqInfo,read,[readQualities],exons)
      else:
         if heuristic_mode:
            if dataset.has_key(id):
               dataset[id].append((currentSeqInfo,read,readQualities,line))
            else:
               dataset[id] = [(currentSeqInfo,read,readQualities,line)]
         else:

            #chromo = seqInfo.chromo_map[id]
            #total_size = seqInfo.chromo_sizes[chromo]

            # determine alignment boundaries
            w_size = settings['half_window_size']
            chromStart_=chromStart-w_size 
            #if chromStart_<1: chromStart_=1
            chromEnd_=chromEnd+w_size
            #if chromEnd_>total_size: chromEnd_=total_size

            currentSeqInfo = id,chromo,strand,chromStart,chromEnd
            if dataset.has_key(id):
               dataset[id].append((currentSeqInfo,read,readQualities))
            else:
               dataset[id] = [(currentSeqInfo,read,readQualities)]


   return dataset
