#!/usr/bin/env python 
# -*- coding: utf-8 -*- 

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Gunnar Raetsch, Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import cPickle
import sys
import pdb
import os
import os.path
import resource

from QPalma.computeSpliceWeights import *
from QPalma.set_param_palma import *
from QPalma.computeSpliceAlignWithQuality import *

from numpy.matlib import mat,zeros,ones,inf
from numpy import inf,mean

from ParaParser import ParaParser,IN_VECTOR

#from QPalma.Lookup import LookupTable
from QPalma.DatasetUtils import parse_bed_format
from QPalma.sequence_utils import SeqSpliceInfo,DataAccessWrapper,reverse_complement,unbracket_seq

import pdb

jp = os.path.join

def cpu():
   return (resource.getrusage(resource.RUSAGE_SELF).ru_utime+\
   resource.getrusage(resource.RUSAGE_SELF).ru_stime) 


class NewWrapper:

   def __init__(self,bed_data):
      self.bed_data = bed_data

   def __len__(self):
      return len(self.bed_data)
      
   def __getitem__(self,key):
      bed_entry  = self.bed_data[key][0]

      currentSeqInfo,read,readQualitiesList,orig_line = bed_entry
      id,chromo,strand,chromStart,chromEnd = currentSeqInfo 

      entry = {}
      entry['id'] = id 
      entry['chr'] = chromo
      entry['pos'] = chromStart
      entry['strand'] = strand
      entry['seq'] = read
      entry['prb'] = readQualitiesList
      return (entry,orig_line.strip())

   def __iter__(self):
      self.counter = 0
      self.size = len(self.bed_data)
      return self

   def next(self):
      if not self.counter < self.size:
         raise StopIteration
      return_val = self.__getitem__(self.bed_data.keys()[self.counter])
      self.counter += 1
      return return_val


class BracketWrapper:
   fields = ['id', 'chr', 'pos', 'strand', 'seq', 'prb']
   fmt_string = "%lu%d%d%s%s%s"

   def __init__(self,filename):
      self.parser = ParaParser(self.fmt_string,self.fields,len(self.fields),IN_VECTOR)
      self.parser.parseFile(filename)

   def __len__(self):
      return self.parser.getSize(IN_VECTOR)
      
   def __getitem__(self,key):
      return self.parser.fetchEntry(key)

   def __iter__(self):
      self.counter = 0
      self.size = self.parser.getSize(IN_VECTOR)
      return self

   def next(self):
      if not self.counter < self.size:
         raise StopIteration
      return_val = self.parser.fetchEntry(self.counter)
      self.counter += 1
      return return_val


class PHeuristic:
   """
   This class wraps the filter which decides whether an alignment found by
   vmatch is spliced an should be then newly aligned using QPalma or not.
   """

   def __init__(self,data_fname,param_fname,result_dir,settings):
      """
      We need a run object holding information about the nr. of support points
      etc.
      """

      self.settings = settings

      start = cpu()
      #self.all_remapped_reads = BracketWrapper(data_fname)
      #bed_data = parse_bed_format('/fml/ag-raetsch/home/fabio/svn/projects/QPalma_release/examples/filtering_tests.bed',settings,False,True)
      bed_data = parse_bed_format(data_fname,settings,False,True)
      self.all_remapped_reads = NewWrapper(bed_data)
      stop = cpu()

      print 'parsed %d reads in %f sec' % (len(self.all_remapped_reads),stop-start)

      self.chromo_range = settings['allowed_fragments']

      accessWrapper = DataAccessWrapper(settings)
      self.seqInfo = SeqSpliceInfo(accessWrapper,settings['allowed_fragments'])

      #start = cpu()
      # get rid of LookupTable
      #self.lt1 = LookupTable(settings)
      #stop = cpu()

      #print 'prefetched sequence and splice data in %f sec' % (stop-start)

      self.result_spliced_fh     = open(jp(result_dir,'filtering_results.spliced'),'w+')
      self.result_unspliced_fh   = open(jp(result_dir,'filtering_results.unspliced'),'w+')

      start = cpu()

      self.data_fname = data_fname

      self.param = cPickle.load(open(param_fname))
      
      # Set the parameters such as limits penalties for the Plifs
      [h,d,a,mmatrix,qualityPlifs] = set_param_palma(self.param,settings)

      self.h = h
      self.d = d
      self.a = a
      self.mmatrix = mmatrix
      self.qualityPlifs = qualityPlifs

      self.prb_offset   = settings['prb_offset']

      # parameters of the heuristics to decide whether the read is spliced
      #self.splice_thresh      = 0.005
      self.splice_thresh      = 0.01
      self.max_intron_size    = settings['max_intron_len']
      self.max_mismatch       = 2 
      #self.splice_stop_thresh = 0.99
      self.splice_stop_thresh = 0.8
      self.spliced_bias       = 0.0

      param_lines = self.printParamLines(param_fname)

      for p_line in param_lines:
         self.result_spliced_fh.write(p_line)
         self.result_unspliced_fh.write(p_line)
   
      lengthSP    = settings['numLengthSuppPoints']
      donSP       = settings['numDonSuppPoints']
      accSP       = settings['numAccSuppPoints']
      mmatrixSP   = settings['matchmatrixRows']*settings['matchmatrixCols']
      numq        = settings['numQualSuppPoints']
      totalQualSP = settings['totalQualSuppPoints']

      currentPhi = zeros((settings['numFeatures'],1))
      currentPhi[0:lengthSP]                                            = mat(h.penalties[:]).reshape(lengthSP,1)
      currentPhi[lengthSP:lengthSP+donSP]                               = mat(d.penalties[:]).reshape(donSP,1)
      currentPhi[lengthSP+donSP:lengthSP+donSP+accSP]                   = mat(a.penalties[:]).reshape(accSP,1)
      currentPhi[lengthSP+donSP+accSP:lengthSP+donSP+accSP+mmatrixSP]   = mmatrix[:]

      totalQualityPenalties = self.param[-totalQualSP:]
      currentPhi[lengthSP+donSP+accSP+mmatrixSP:]                       = totalQualityPenalties[:]
      self.currentPhi = currentPhi

      # total time spend for get seq and scores
      self.get_time  = 0.0
      self.calcAlignmentScoreTime = 0.0
      self.alternativeScoresTime = 0.0

      # some variabes storing the run time of the steps we perform
      self.count_time = 0.0
      self.main_loop = 0.0
      self.splice_site_time = 0.0
      self.computeSpliceAlignWithQualityTime = 0.0
      self.computeSpliceWeightsTime = 0.0
      self.DotProdTime = 0.0
      self.array_stuff = 0.0
      stop = cpu()

      self.init_time = stop-start


   def printParamLines(self,param_fname):
      param_lines = [\
      ('%f','splice_thresh',self.splice_thresh),\
      ('%d','max_intron_size',self.max_intron_size),\
      ('%d','max_mismatch',self.max_mismatch),\
      ('%f','splice_stop_thresh',self.splice_stop_thresh),\
      ('%f','spliced_bias',self.spliced_bias)]

      param_lines = [('# %s: '+form+'\n')%(name,val) for form,name,val in param_lines]
      param_lines.append('# data: %s\n'%self.data_fname)
      param_lines.append('# param: %s\n'%param_fname)
      return param_lines


   def filter(self):
      """
      This method...
      """

      ctr,spliced_ctr,unspliced_ctr    = 0,0,0

      print 'Starting filtering...'
      _start = cpu()

      
      for location,original_line in self.all_remapped_reads:
         
         #pdb.set_trace()
         
         if ctr % 1000 == 0:
            print ctr

         id       = int(location['id'])
         chromo   = location['chr']
         pos      = location['pos']
         strand   = location['strand']
         seq      = location['seq'].lower()
         prb      = location['prb']

         if not chromo in self.chromo_range:
            continue

         unb_seq = unbracket_seq(seq)

         # forgot to do this
         if strand == '-':
            unb_seq = reverse_complement(unb_seq)
            seq = reverse_complement(seq)

         effective_len = len(unb_seq)

         genomicSeq_start  = pos
         genomicSeq_stop   = pos+effective_len

         start = cpu()
         #currentDNASeq, currentAcc, currentDon = self.lt1.get_seq_and_scores(chromo,strand,genomicSeq_start,genomicSeq_stop)
         currentDNASeq, currentAcc, currentDon = self.seqInfo.get_seq_and_scores(chromo,strand,genomicSeq_start,genomicSeq_stop)
         stop = cpu()

         self.get_time += stop-start

         dna            = currentDNASeq
         exons          = zeros((2,1))
         exons[0,0]     = 0
         exons[1,0]     = effective_len
         est            = unb_seq
         original_est   = seq
         quality        = prb

         currentVMatchAlignment = dna, exons, est, original_est, quality,\
         currentAcc, currentDon

         #pdb.set_trace()
         
         #try:
         alternativeAlignmentScores = self.calcAlternativeAlignments(location)
         #except:
            #print 'alternativeAlignmentScores = []'
         #   alternativeAlignmentScores = []

         if alternativeAlignmentScores == []:
             # no alignment necessary
             maxAlternativeAlignmentScore = -inf
             vMatchScore = 0.0 
         else:
             maxAlternativeAlignmentScore = max(alternativeAlignmentScores)
             # compute alignment for vmatch unspliced read
             vMatchScore = self.calcAlignmentScore(currentVMatchAlignment)
             #print 'vMatchScore/alternativeScore: %f %f %s' % (vMatchScore,maxAlternativeAlignmentScore,strand)
             
         start = cpu()
         
         # Seems that according to our learned parameters VMatch found a good
         # alignment of the current read
         if maxAlternativeAlignmentScore < vMatchScore:
            self.result_unspliced_fh.write(original_line+'\n') 
            unspliced_ctr += 1

         # We found an alternative alignment considering splice sites that scores
         # higher than the VMatch alignment
         else:
            self.result_spliced_fh.write(original_line+'\n')
            spliced_ctr += 1

         stop = cpu()
         self.count_time = stop-start
         ctr += 1

      _stop = cpu()
      self.main_loop = _stop-_start

      print 'Unspliced/Spliced: %d %d'%(unspliced_ctr,spliced_ctr)


   def findHighestScoringSpliceSites(self, currentAcc, currentDon, DNA, start_end, read_size, splice_thresh):

      def signum(a):
          if a>0: 
              return 1
      	  elif a<0:
              return -1
          else:
              return 0

      (read_start, read_end)=start_end ;
      proximal_acc   = []
      for idx in xrange(1, read_start+read_size/2):
          if currentAcc[idx]>= splice_thresh:
            proximal_acc.append((idx,currentAcc[idx]))

      proximal_acc.sort(lambda x,y: signum(x[1]-y[1])) 
      proximal_acc=proximal_acc[-2:]

      distal_acc   = []
      for idx in xrange(read_end, len(currentAcc)):
          if currentAcc[idx]>= splice_thresh and idx+read_size<len(currentAcc):
            distal_acc.append((idx, currentAcc[idx], DNA[idx+1:idx+read_size+1]))

      #distal_acc.sort(lambda x,y: signum(x[1]-y[1])) 
      #distal_acc=distal_acc[-2:]

      proximal_don   = []
      for idx in xrange(read_start+read_size/2, len(currentDon)):
         if currentDon[idx] >= splice_thresh:
            proximal_don.append((idx, currentDon[idx]))

      proximal_don.sort(lambda x,y: signum(x[1]-y[1]))
      proximal_don=proximal_don[-2:]

      distal_don   = []
      for idx in xrange(1, read_start):
         if currentDon[idx] >= splice_thresh and idx>read_size:
            distal_don.append((idx, currentDon[idx], DNA[idx-read_size:idx]))

      distal_don.sort(lambda x,y: y[0]-x[0])
      #distal_don=distal_don[-2:]

      return proximal_acc,proximal_don,distal_acc,distal_don


   def calcAlternativeAlignments(self,location):
      """
      Given an alignment proposed by Vmatch this function calculates possible
      alternative alignments taking into account for example matched
      donor/acceptor positions.
      """

      id       = location['id']
      chromo_id   =  location['chr']
      chromo =  self.seqInfo.chromo_map[chromo_id] 
      pos      = location['pos']
      strand   = location['strand']
      original_est = location['seq']
      quality      = location['prb']
      #pdb.set_trace() ;
      #quality        = map(lambda x:ord(x)-self.prb_offset,quality)
      
      original_est = original_est.lower()
      est = unbracket_seq(original_est)
      effective_len = len(est)

      genomicSeq_start  = pos - self.max_intron_size
      if genomicSeq_start<0:
          genomicSeq_start=0 ;
      genomicSeq_stop   = pos + self.max_intron_size + len(est)
      if genomicSeq_stop > self.seqInfo.chromo_sizes[chromo]:
          genomicSeq_stop = self.seqInfo.chromo_sizes[chromo] ;
          
      if strand == '-':
         est = reverse_complement(est)
         original_est = reverse_complement(original_est)

      start = cpu()
      #currentDNASeq, currentAcc, currentDon = self.lt1.get_seq_and_scores(chromo_id, strand, genomicSeq_start, genomicSeq_stop)
      currentDNASeq, currentAcc, currentDon = self.seqInfo.get_seq_and_scores(chromo_id, strand, genomicSeq_start, genomicSeq_stop)
      stop = cpu()
      self.get_time += stop-start
      dna   = currentDNASeq

      proximal_acc,proximal_don,distal_acc,distal_don = self.findHighestScoringSpliceSites(currentAcc, currentDon, dna, (pos, pos+len(est)), len(est), self.splice_thresh)
	
      alternativeScores = []

      #pdb.set_trace()
      
      # inlined
      h = self.h
      d = self.d
      a = self.a
      mmatrix = self.mmatrix
      qualityPlifs = self.qualityPlifs
      # inlined

      # find an intron on the 3' end
      _start = cpu()
      for (don_pos,don_score) in proximal_don:
          DonorScore = calculatePlif(d, [don_score])[0]

          for (acc_pos,acc_score,acc_dna) in distal_acc:

              IntronScore = calculatePlif(h, [acc_pos-don_pos])[0] 
              AcceptorScore = calculatePlif(a, [acc_score])[0] 
          
              #print 'don splice: ', (don_pos,don_score), (acc_pos,acc_score,acc_dna), (DonorScore,IntronScore,AcceptorScore)

              # construct a new "original_est"
              original_est_cut=''
              
              est_ptr=0
              dna_ptr = pos-genomicSeq_start #self.max_intron_size
              ptr=0
              acc_dna_ptr=0
              num_mismatch = 0
              
              while ptr<len(original_est):
                  #print acc_dna_ptr,len(acc_dna),acc_pos,don_pos
                  
                  if original_est[ptr]=='[':
                      dnaletter=original_est[ptr+1]
                      estletter=original_est[ptr+2]
                      if dna_ptr < don_pos:
                          original_est_cut+=original_est[ptr:ptr+4]
                          num_mismatch += 1
                      else:
                          if acc_dna[acc_dna_ptr]==estletter:
                              original_est_cut += estletter # EST letter
                          else:
                              original_est_cut += '['+acc_dna[acc_dna_ptr]+estletter+']' # EST letter
                              num_mismatch += 1
                              #print '['+acc_dna[acc_dna_ptr]+estletter+']' 
                          acc_dna_ptr+=1 
                      ptr+=4 
                  else:
                      dnaletter=original_est[ptr]
                      estletter=dnaletter
                
                      if dna_ptr < don_pos:
                          original_est_cut+=estletter # EST letter
                      else:
                          if acc_dna[acc_dna_ptr]==estletter:
                              original_est_cut += estletter # EST letter
                          else:
                              num_mismatch += 1
                              original_est_cut += '['+acc_dna[acc_dna_ptr]+estletter+']' # EST letter
                              #print '('+acc_dna[acc_dna_ptr]+estletter+')' 
                          acc_dna_ptr+=1 
                          
                      ptr+=1

                  if estletter=='-':
                      dna_ptr+=1 
                  elif dnaletter=='-':
                      est_ptr+=1
                  else:
                      dna_ptr+=1 
                      est_ptr+=1
              if num_mismatch>self.max_mismatch:
                  continue
                         
              assert(dna_ptr<=len(dna))
              assert(est_ptr<=len(est))

              #print original_est, original_est_cut              

              score = computeSpliceAlignScoreWithQuality(original_est_cut, quality, qualityPlifs, self.settings, self.currentPhi)
              score += AcceptorScore + IntronScore + DonorScore + self.spliced_bias 
         
              alternativeScores.append(score)

              #print "acc_score=%1.2f" % acc_score
              
              if acc_score>=self.splice_stop_thresh:
                  break
              
      #pdb.set_trace()

      _stop = cpu()
      self.alternativeScoresTime += _stop-_start

      # find an intron on the 5' end 
      _start = cpu()
      for (acc_pos,acc_score) in proximal_acc:

          AcceptorScore = calculatePlif(a, [acc_score])[0]
          
          for (don_pos,don_score,don_dna) in distal_don:

              DonorScore = calculatePlif(d, [don_score])[0]
              IntronScore = calculatePlif(h, [acc_pos-don_pos])[0] 
          
              #print 'acc splice: ', (don_pos,don_score,don_dna), (acc_pos,acc_score), (DonorScore,IntronScore,AcceptorScore)

              # construct a new "original_est"
              original_est_cut=''
              
              est_ptr=0
              dna_ptr = pos-genomicSeq_start #self.max_intron_size
              ptr=0
              num_mismatch = 0
              don_dna_ptr=len(don_dna)-(acc_pos-dna_ptr)-1
              while ptr<len(original_est):
                  
                  if original_est[ptr]=='[':
                      dnaletter=original_est[ptr+1]
                      estletter=original_est[ptr+2]
                      if dna_ptr > acc_pos:
                          original_est_cut+=original_est[ptr:ptr+4] 
                          num_mismatch += 1
                      else:
                          if don_dna[don_dna_ptr]==estletter:
                              original_est_cut += estletter # EST letter
                          else:
                              original_est_cut += '['+don_dna[don_dna_ptr]+estletter+']' # EST letter
                              num_mismatch += 1
                              #print '['+don_dna[don_dna_ptr]+estletter+']' 
                          don_dna_ptr+=1 
                      ptr+=4 
                  else:
                      dnaletter=original_est[ptr]
                      estletter=dnaletter
                
                      if dna_ptr > acc_pos:
                          original_est_cut+=estletter # EST letter
                      else:
                          if don_dna[don_dna_ptr]==estletter:
                              original_est_cut += estletter # EST letter
                          else:
                              original_est_cut += '['+don_dna[don_dna_ptr]+estletter+']' # EST letter
                              num_mismatch += 1
                              #print '('+don_dna[don_dna_ptr]+estletter+')' 
                          don_dna_ptr+=1 
                          
                      ptr+=1

                  if estletter=='-':
                      dna_ptr+=1 
                  elif dnaletter=='-':
                      est_ptr+=1
                  else:
                      dna_ptr+=1 
                      est_ptr+=1
                         
              if num_mismatch>self.max_mismatch:
                  continue
              assert(dna_ptr<=len(dna))
              assert(est_ptr<=len(est))

              score = computeSpliceAlignScoreWithQuality(original_est_cut, quality, qualityPlifs, self.settings, self.currentPhi)
              score += AcceptorScore + IntronScore + DonorScore + self.spliced_bias
         
              alternativeScores.append(score)

              #print "don_score=%1.2f" % don_score

              if don_score>=self.splice_stop_thresh:
                  break
              
      _stop = cpu()
      self.alternativeScoresTime += _stop-_start

      return alternativeScores


   def calcAlignmentScore(self,alignment):
      """
      Given an alignment (dna,exons,est) and the current parameter for QPalma
      this function calculates the dot product of the feature representation of
      the alignment and the parameter vector i.e the alignment score. 
      """

      start = cpu()

      # Lets start calculation
      dna, exons, est, original_est, quality, acc_supp, don_supp = alignment
      score = computeSpliceAlignScoreWithQuality(original_est, quality, self.qualityPlifs, self.settings, self.currentPhi)

      stop = cpu()
      self.calcAlignmentScoreTime += stop-start
      
      return score 
