# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import math
import pdb

from numpy.matlib import zeros,isinf

import QPalmaDP


class Plf: 
   """
   means piecewise linear function
   """

   def __init__(self,num=None):
      if num == None:
         self.len = 0
         self.limits = []
         self.penalties = []
         self.transform = ''
         self.name = ''
         self.max_len = 0
         self.min_len = 0
      else:
         self.len = num
         self.limits = linspace(0,40,num) 
         self.penalties = [0]*num
         self.transform = ''
         self.name = ''
         self.max_len = 0
         self.min_len = 0
         

   def convert2SWIG(self):
      ps = QPalmaDP.penalty_struct()

      ps.len = len(self.limits)
   
      ps.limits = QPalmaDP.createDoubleArrayFromList(self.limits)
      ps.penalties = QPalmaDP.createDoubleArrayFromList(self.penalties)

      ps.max_len = self.max_len
      ps.min_len = self.min_len
      ps.transform = 0
      ps.name = self.name

      return ps


def base_coord(b1,b2):
   b1 = b1.lower()
   b2 = b2.lower()
   assert b1 in ['a','c','g','t','n']
   assert b2 in ['a','c','g','t','n']
   cb = {'a':0, 'c':1, 'g':2, 't':3, 'n':4}

   b1 = cb[b1]
   b2 = cb[b2]

   return b1*5+b2

def linspace(a,b,n):
   intervalLength = b-a
   stepSize = 1.0*intervalLength / (n-1)
   
   interval = [0]*n
   interval[0] = a
   interval[-1] = b
   for i in range(1,n-1):
      interval[i] = a+(i*stepSize)

   return interval

def logspace(a,b,n):
   interval = [0]*n
   begin = 10.0**a
   end = 10.0**b
   intervalSize = 1.0*(b-a)/(n-1)
   interval[0] = begin
   interval[-1] = end

   for i in range(1,n-1):
      interval[i] = 10**(a+i*intervalSize)

   return interval

def log10(x):
   return math.log(x)/math.log(10)


def penalty_lookup_new(penalty_struct, value):

   limits = penalty_struct.limits
   penalties = penalty_struct.penalties

   if penalty_struct.transform == 'log':
      value = math.log(value)

   elif penalty_struct.transform == 'log(+3)':
      value = math.log(value+3)

   elif penalty_struct.transform == 'log(+1)':
      value = math.log(value+1)

   elif penalty_struct.transform == '':
      pass

   elif penalty_struct.transform == '(+3)':
      value += 3
         
   elif penalty_struct.transform == 'mod3':
     value = value%3

   else:
      assert False,'unknown penalty transform'

   count = len([elem for elem in limits if elem <= value])

   if count == 0:
      pen = penalties[0]
   elif count == len(limits):
      pen=penalties[-1]
   else:
      pen = (penalties[count]*(value-limits[count-1]) + penalties[count-1]\
	  *(limits[count]-value)) / (limits[count]-limits[count-1])

   return pen


def compute_donacc(donor_supp, acceptor_supp, d, a):

   assert(len(donor_supp)==len(acceptor_supp))
   size = len(donor_supp)
  
   donor    = [0.0]*size
   acceptor = [0.0]*size

   for idx in range(size):
     if isinf(donor_supp[idx]):
       donor[idx] = donor_supp[idx]
     else:
       donor[idx] = penalty_lookup_new(d, donor_supp[idx])

     if isinf(acceptor_supp[idx]):
       acceptor[idx] = acceptor_supp[idx]
     else:
       acceptor[idx] = penalty_lookup_new(a,acceptor_supp[idx])

   return donor,acceptor


