# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

#
# This file contains the main interface to the QPalma pipeline.
#

import os 
import re
import os.path 
import sys 
import pdb

jp = os.path.join

try:
   import SIQP_CVXOPT
   has_cvxopt = True
except:
   has_cvxopt = False

try:
   import SIQP_MOSEK
   has_mosek = True
except:
   has_mosek = False


def parseSettingsFile(filename):
   """
   This function parse all key value pairs from the given filename
   """
   settings = {}
   for line in open(filename):
      if (not line.strip()) or line.startswith('#'):
         pass
      else:
         line = line.strip()
         line = re.sub('\s', '', line)  
         key, val = line.split('=')

         settings[key] = val

   return settings


def makeSettings(settings):
   """
   
   """

   # addd checking support!!
   settings['allowed_fragments'] = eval(settings['allowed_fragments'])

   try:
      settings['pos_strand_prefetching_enabled'] = eval(settings['pos_strand_prefetching_enabled'])
   except:
      settings['pos_strand_prefetching_enabled'] = True

   try:
      settings['neg_strand_prefetching_enabled'] = eval(settings['neg_strand_prefetching_enabled'])
   except:
      settings['neg_strand_prefetching_enabled'] = True

   TF = [True,False]

   # convert and check all boolean parameters
   for parameter in ['enable_quality_scores', 'enable_intron_length', 'enable_splice_scores', 'remove_duplicate_scores', 'print_matrix', 'perform_checks', 'drop_unspliced']:
      try:
         settings[parameter]   = eval(settings[parameter])
         assert settings[parameter] in TF
      except:
         print "Unexpected error:", sys.exc_info()[0]
         print 'The value of %s has to be either True or False!'%parameter


   if settings['enable_quality_scores']:
      matchmatrixRows         = 6
      matchmatrixCols         = 1
   else:
      matchmatrixRows         = 6
      matchmatrixCols         = 6

   settings['matchmatrixRows'] = matchmatrixRows
   settings['matchmatrixCols'] = matchmatrixCols

   # now convert and check all integer parameters
   for parameter in ['numAccSuppPoints', 'numConstraintsPerRound',\
   'numDonSuppPoints', 'numLengthSuppPoints', 'numQualPlifs', 'numQualSuppPoints', 'iter_steps',\
   'max_intron_len', 'max_qual', 'min_intron_len', 'min_qual', 'prb_offset',\
   'half_window_size']:
      try:
         settings[parameter] = int(settings[parameter])
      except:
         print 'Error: %s has to be an integer!'%parameter

   # now convert and check all float parameters
   for parameter in ['max_svm_score','min_svm_score','C']:
      try:
         settings[parameter] = float(settings[parameter])
      except:
         print 'Error: %s has to be an float!'%parameter

   settings['numFeatures'] = settings['numLengthSuppPoints'] + settings['numAccSuppPoints'] + settings['numDonSuppPoints']\
   + settings['matchmatrixCols']*settings['matchmatrixRows'] + settings['numQualPlifs']*settings['numQualSuppPoints']

   settings['totalQualSuppPoints'] = settings['numQualPlifs']*settings['numQualSuppPoints']

   # we now support the expansion of environment variables
   for (key, val,) in settings.items():
      if key.endswith('_fn') or key.endswith('_dir') or key.endswith('_loc'):
         print os.path.expandvars(val)
         settings[key] = os.path.expandvars(val)

   if not has_mosek and not has_cvxopt:
      print 'Error: There is no optimization toolkit available!'
      sys.exit(1)

   settings['optimizer'] = settings['optimizer'].lower()
   assert settings['optimizer'] in ['auto','cvxopt','mosek']

   if settings['optimizer'] == 'cvxopt':
      if has_cvxopt:
         settings['solver_id'] = 0
      else:
         print 'Warning: You set \'optimizer\' to cvxopt but this package is not installed'
         print 'I am setting \'optimizer\' to \'auto\' to use a different optimizer'
         settings['optimizer'] = 'auto'

   if settings['optimizer'] == 'mosek':
      if has_mosek:
         settings['solver_id'] = 1
      else:
         print 'Warning: You set \'optimizer\' to mosek but this package is not installed'
         print 'I am setting \'optimizer\' to \'auto\' to use a different optimizer'
         settings['optimizer'] = 'auto'

   if settings['optimizer'] == 'auto':
      if has_mosek:
         settings['solver_id'] = 1
      elif has_cvxopt:
         settings['solver_id'] = 0
      else:
         assert False

   return settings


def checkSettings(settings):
   for (key, val,) in settings.items():
      if key.endswith('_fn') or key.endswith('_dir') or key.endswith('_loc'):
         assert os.path.exists(val), 'Error: Path/File "%s" with value "%s" does not seem to exist!' % (key,val)

   return True


def parseSettings(filename):
   settings = parseSettingsFile(filename)
   settings = makeSettings(settings)
   assert checkSettings(settings)
   return settings

if __name__ == '__main__':
   print parseSettings(sys.argv[1])
