# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import numpy
from numpy.matlib import mat,zeros,ones,inf
import pdb
from Plif import Plf,base_coord,linspace

import sequence_utils

#mytime=numpy.zeros(4) ;

def  computeSpliceAlign(dna, exons, est, original_est, quality, qualityPlifs, settings):
   """
   Exonpos: Anfang Exon bis 1 rechts neben Ende Exon
   Daraus folgt: SpliceposAnfang ist gleich Ende Exon von Vorgaenger
   (ausser dem letzten und ersten)
    und Ende Splicestelle ist gleich Anfang Exon vom naechsten, da Splicestelleende notiert 1 rechts neben Splciestellenende
   bsp: 
   cccccA1cccccE1cccccA2 ... AmcccccEmcccccccc
   cccccXXcccccA1cccccE1 ... Em-1cccXXcccccccc
   """
   import time
   start = time.clock()
   
   SpliceAlign = []

   #mytime[0] = mytime[0] + (time.clock() - start) ;

   if not (exons.shape == (1,2)):
      numberOfExons = exons.shape[0]
      exonSizes = [-1]*numberOfExons
      for idx in range(numberOfExons):
          exonSizes[idx] = exons[idx,1] - exons[idx,0]

      # SpliceAlign vector describes alignment: 
      # 1:donorpos, 3:intron 2:acceptorpos, 0:exon, 4: dangling end

      if exons[0,0] > 0:
         SpliceAlign.extend([4]*(exons[0,0]))

      for idx in range(numberOfExons):
         exonLength = exonSizes[idx]
         SpliceAlign.extend([0]*exonLength)
         
         if idx < numberOfExons-1:
            intronLength = exons[idx+1,0] - exons[idx,1]
            SpliceAlign.extend([1]+[3]*(intronLength-2)+[2])

      if len(dna) > exons[-1,1]:
         SpliceAlign.extend([4]*(len(dna)-exons[-1,1]))

   elif exons.shape == (1,2):

      numberOfExons = 1
      exonSizes = [-1]*numberOfExons

      exonSizes[0] = exons[0,1] - exons[0,0]
      #exonSizes[1] = exons[1,1] - exons[1,0]

      # SpliceAlign vector describes alignment: 
      # 1:donorpos, 3:intron 2:acceptorpos, 0:exon, 4: dangling end

      if exons[0,0] > 0:
         SpliceAlign.extend([4]*(exons[0,0]))

      for idx in range(numberOfExons):
         exonLength = exonSizes[idx]
         SpliceAlign.extend([0]*exonLength)
         
         if idx < numberOfExons-1:
            intronLength = exons[idx+1,0] - exons[idx,1]
            SpliceAlign.extend([1]+[3]*(intronLength-2)+[2])

      if len(dna) > exons[-1,1]:
         SpliceAlign.extend([4]*(len(dna)-exons[-1,1]))

      #exonSize = exons[0,1] - exons[0,0]

      #if exons[0,0] > 0:
      #   SpliceAlign.extend([4]*(exons[0,0]))

      #SpliceAlign.extend([0]*(exonSize))

      #if len(dna) > exons[0,1]:
      #   SpliceAlign.extend([4]*(len(dna)-exons[0,1]))

   else:
      assert(False) 
      pass

   assert len(SpliceAlign) == len(dna), pdb.set_trace()

   #  
   # start of label feature generation
   #  

   #mytime[1] = mytime[1] + (time.clock() - start) ;

   sizeMatchmatrix = settings['matchmatrixRows']*settings['matchmatrixCols'] #
   # counts the occurences of a,c,g,t,n in this order

   # counts the occurrences of a,c,g,t,n with their quality scores
   if settings['enable_quality_scores']:
       trueWeightQualityMat = [0.0]*5 # 5 rows
       for row in range(5):
           trueWeightQualityMat[row] = [0.0]*6 # 6 col per row

       for row in range(5):
           for col in range(6):
               trueWeightQualityMat[row][col] = [0.0]*settings['numQualSuppPoints'] # supp. points per plif
       
   dna_part = []
   est_part = []

   orig_pos = 0
   while True:
      if not orig_pos < len(original_est):
         break

      orig_char = original_est[orig_pos]

      if orig_char == '[':
         dna_part.append( original_est[orig_pos+1] )
         est_part.append( original_est[orig_pos+2] )
         orig_pos += 4
      else:
         dna_part.append( orig_char )
         est_part.append( orig_char )
         orig_pos += 1

   for orig_char in dna_part:
      assert orig_char in sequence_utils.alphabet, pdb.set_trace()

   for orig_char in est_part:
      assert orig_char in sequence_utils.alphabet, pdb.set_trace()

   #mytime[2] = mytime[2] + (time.clock() - start) ;

   trueWeightMatch = zeros((settings['matchmatrixRows']*settings['matchmatrixCols'],1)) # Scorematrix fuer Wahrheit

   map = {'-':0, 'a':1, 'c':2, 'g':3, 't':4, 'n':5}

   if settings['enable_quality_scores']:
       ctr = 0
       quality_part = []
       for elem in est_part:
           if elem == '-':
               quality_part.append(-inf)
           else:
               if ctr>=len(quality):
                   quality_part.append(quality[0]) # hack
               else:
                   quality_part.append(quality[ctr])
                   ctr += 1
       assert len(est_part) == len(quality_part)

   if not settings['enable_quality_scores']:
       for idx in range(len(dna_part)):
           dnanum = map[dna_part[idx]]
           estnum = map[est_part[idx]]

           if estnum == 0:
               trueWeightMatch[dnanum] += 1.0
               
           else:
               cur_elem = (estnum-1)*6+dnanum
               trueWeightMatch[cur_elem] += 1.0

           trueWeightQuality = zeros((0,1))

   if settings['enable_quality_scores']:
       for idx in range(len(dna_part)):
           dnanum = map[dna_part[idx]]
           estnum = map[est_part[idx]]
           value = quality_part[idx]
           
           if estnum == 0:
               trueWeightMatch[dnanum] += 1.0
               assert value == -inf
               
           else:
               assert value != -inf
               cur_plf = qualityPlifs[(estnum-1)*6+dnanum]
               
               Lower = len([elem for elem in cur_plf.limits if elem <= value])
               
               if Lower == 0:
                   trueWeightQualityMat[estnum-1][dnanum][0] += 1
               elif Lower == len(cur_plf.limits):
                   trueWeightQualityMat[estnum-1][dnanum][-1] += 1
               else:
                   # because we count from 0 in python
                   Lower -= 1
                   Upper = Lower+1 ; # x-werte bleiben fest
                   #print value,Lower,Upper
                   weightup  = 1.0*(value - cur_plf.limits[Lower]) / (cur_plf.limits[Upper] - cur_plf.limits[Lower])
                   weightlow = 1.0*(cur_plf.limits[Upper] - value) / (cur_plf.limits[Upper] - cur_plf.limits[Lower])
                   
                   trueWeightQualityMat[estnum-1][dnanum][Upper] += weightup
                   trueWeightQualityMat[estnum-1][dnanum][Lower] += weightlow

       trueWeightQuality = numpy.array(trueWeightQualityMat).reshape((5*6*settings['numQualSuppPoints'],1)) ;
       if False:
           trueWeightQuality = zeros((settings['numQualSuppPoints']*settings['numQualPlifs'],1))
           ctr = 0
           for row in range(5):
               for col in range(6):
                   for p_idx in range(settings['numQualSuppPoints']):
                       #print ctr, row, col, p_idx
                       trueWeightQuality[ctr] = trueWeightQualityMat[row][col][p_idx]
                       assert trueWeightQuality[ctr] == trueWeightQuality2[ctr]
                       ctr += 1
               #pdb.set_trace()

   #mytime[3] = mytime[3] + (time.clock() - start) ;
   #pdb.set_trace()

   #pdb.set_trace()
   #featureSum = 0
   #featureSum += sum([e for p,e in enumerate(trueWeightQuality.flatten().tolist()[0]) if e != 0.0])
   #featureSum += sum(trueWeightMatch.flatten().tolist()[0])

   #assert int(featureSum) == len(est_part),pdb.set_trace()
   # 'feature sum is not equal read size!!'

   dna_orig=dna[exons[0,0]:exons[0,1]] ;
   for idx in xrange(1, exons.shape[0]):
       dna_orig += dna[exons[idx,0]:exons[idx,1]]

   dna_calc = "".join(dna_part)

   #assert dna_orig == dna_calc, pdb.set_trace()

   #print mytime
   
   return SpliceAlign, trueWeightMatch, trueWeightQuality, dna_calc


def  computeSpliceAlignScoreWithQuality(original_est, quality, qualityPlifs, settings, weightvector):
   """
   """

   lengthSP    = settings['numLengthSuppPoints']
   donSP       = settings['numDonSuppPoints']
   accSP       = settings['numAccSuppPoints']
   mmatrixSP   = settings['matchmatrixRows']*settings['matchmatrixCols']
   numq        = settings['numQualSuppPoints']*5*6

   lengthP = weightvector[0:lengthSP]
   donP = weightvector[lengthSP:lengthSP+donSP]
   accP = weightvector[donSP+lengthSP:donSP+lengthSP+accSP]
   mmatrixP = weightvector[accSP+donSP+lengthSP:accSP+donSP+lengthSP+mmatrixSP]
   numqP = weightvector[accSP+donSP+lengthSP+mmatrixSP:accSP+donSP+lengthSP+mmatrixSP+numq]
   numQualSuppPoints=settings['numQualSuppPoints']
   
   map = {'-':0, 'a':1, 'c':2, 'g':3, 't':4, 'n':5, '[': 10, ']':11 }
   original_est_mapped = [ map[e] for e in original_est ]

   if True:
       score=0.0
       est_ptr=0
       dna_ptr=0
       ptr=0 
       while ptr<len(original_est_mapped):
           
           if original_est_mapped[ptr]==10: #'[':
               dnaletter=original_est_mapped[ptr+1]
               estletter=original_est_mapped[ptr+2]
               ptr+=4 
           else:
               dnaletter=original_est_mapped[ptr]
               estletter=dnaletter
               ptr+=1

           if estletter==0:# '-':
               score+= mmatrixP[dna_letter]
           else:
               value = quality[est_ptr]
               cur_plf = qualityPlifs[(estletter-1)*6+dnaletter]
               
               Lower = 0
               for elem in cur_plf.limits:
                   if elem<=value:
                       Lower+=1 
               
               if Lower == 0:
                   score += numqP[(estletter-1)*6*numQualSuppPoints + dnaletter*numQualSuppPoints+0] ;
                   
               elif Lower == len(cur_plf.limits):
                   score += numqP[(estletter-1)*6*numQualSuppPoints + dnaletter*numQualSuppPoints + numQualSuppPoints-1] ;

               else:
                   # because we count from 0 in python
                   Lower -= 1
                   Upper = Lower+1 ; # x-werte bleiben fest
                   weightup  = 1.0*(value - cur_plf.limits[Lower]) / (cur_plf.limits[Upper] - cur_plf.limits[Lower])
                   weightlow = 1.0*(cur_plf.limits[Upper] - value) / (cur_plf.limits[Upper] - cur_plf.limits[Lower])
                   
                   score += numqP[(estletter-1)*6*numQualSuppPoints + dnaletter*numQualSuppPoints+Upper]*weightup ;
                   score += numqP[(estletter-1)*6*numQualSuppPoints + dnaletter*numQualSuppPoints+Lower]*weightlow ;
                                  
           if estletter==0:# '-':
               dna_ptr+=1 
           elif dnaletter==0:#'-':
               est_ptr+=1
           else:
               dna_ptr+=1 
               est_ptr+=1
               
   return score
