#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os.path
import csv
import re
from PyGff import *
import cPickle
import copy

def parse_gff(gff_filename):
   gff_fid = open(gff_filename)
   reader = csv.reader(gff_fid, delimiter='\t', quoting=csv.QUOTE_NONE)

   allGenes = {}
   allGenesA = []
   currentGene = None

   rx = re.compile('ID=[^;]*;',re.DOTALL)

   for row in reader:
      assert len(row) == 9
      chr = row[0]
      id = row[2]
      start = int(row[3])
      stop = int(row[4])
      strand = row[6]
      desc = row[8]

      if id == 'chromosome':
         continue

      if id == 'gene':
         if currentGene != None:
            allGenes[currentGene.id] = currentGene
            allGenesA.append(currentGene)
         
         desc = rx.search(desc).group()[3:-1]
         #print desc
         currentGene = Gene(chr,start,stop,strand,desc)

      elif id == 'five_prime_UTR':
         pass

      elif id == 'three_prime_UTR':
         pass

      elif id == 'mRNA':
         pass

      elif id == 'exon':
         assert currentGene != None
         currentGene.addExon(start,stop)

      elif id == 'CDS':
         pass

      elif id == 'ncRNA':
         pass

      elif id == 'pseudogenic_exon':
         pass

      elif id == 'pseudogenic_transcript':
         pass

      elif id == 'miRNA':
         pass

      elif id == 'rRNA':
         pass

      elif id == 'snoRNA':
         pass

      elif id == 'snRNA':
         pass

      elif id == 'tRNA':
         pass

      elif id == 'pseudogene':
         if currentGene != None:
            allGenes[currentGene.id] = currentGene
            allGenesA.append(currentGene)
            currentGene = None
      else:
         assert False, 'Error: Unknown identifier \'%s\'' % id

   if currentGene != None:
      allGenes[currentGene.id] = currentGene

   return allGenes# ,allGenesA


def createGffPickle(annotFile,pickleFile):
   #gff_fid = open(annotFile)
   pickle_fid = open(pickleFile,'w+')
   allGenes = parse_gff(annotFile)
   #for key,val in allGenes.iteritems():
      #print key
   cPickle.dump(allGenes,pickle_fid)
      
if __name__ == '__main__':
   assert len(sys.argv) >= 3
   annotFile = sys.argv[1]
   pickleFile = sys.argv[2]
   assert os.path.exists(annotFile)
   assert not os.path.exists(pickleFile)

   createGffPickle(annotFile,pickleFile)
