#!/bin/sh
#

PROG=`basename $0`
DIR=`dirname $0`


. ${DIR}/../../bin/qpalma_config.sh

. $QPALMA_BIN_PATH/qpalma_header.sh $0 $*

set -e

if [ "$#" = "0" ];
then
	echo usage $0 \<orig_sample\> \<ref_seq\> \<subsample_size\> \<error_model\> \<quality_sample\>
	exit -1
fi

aligned_reads=`tempfile`
unmapped_reads=`tempfile`

sub_sample=`tempfile`
gmapper_index_dir=`mktemp -d`
random_index_dir=0
orig_sample_file=$1
ref_seq=$2
subsample_size=$3
error_model=$4
quality_sample=$5

echo python ${QPALMA_TOOLS_PATH}/errormodel/subsample_fastq.py $orig_sample_file $subsample_size $sub_sample
python ${QPALMA_TOOLS_PATH}/sequences/subsample_fastq.py $orig_sample_file $subsample_size $sub_sample
cat $sub_sample | grep "@" -B 1 | grep -v -e '^--' -e "^@" -e "^+"  > $quality_sample

### create indices
if [ ! -f $ref_seq.cid ]
then
    echo "creating index for $ref_seq ..."
    cp $ref_seq $gmapper_index_dir/
    $PALMAPPER_PATH/pmindex -v -i $gmapper_index_dir/`basename $ref_seq`
    random_index_dir=1
   else
    gmapper_index_dir=`dirname $ref_seq`
fi

### align
echo ${PALMAPPER_PATH}/palmapper -q $sub_sample -i $gmapper_index_dir/`basename $ref_seq` -o $aligned_reads -M 6 -G 2 -E 6 -u $unmapped_reads -f bedx -z 5 -seed-hit-cancel-threshold 10000 -index-precache
${PALMAPPER_PATH}/palmapper -q $sub_sample -i $gmapper_index_dir/`basename $ref_seq` -o $aligned_reads -M 6 -G 2 -E 6 -u $unmapped_reads -f bedx -z 5 -seed-hit-cancel-threshold 10000 -index-precache

python ${QPALMA_TOOLS_PATH}/errormodel/mismatch.py $aligned_reads $error_model 8

rm -f $aligned_reads $unmapped_reads $sub_sample
if [ $random_index_dir -gt 0 ]
then
    if [ "$gmapper_index_dir" != "`dirname $ref_seq`" ]
    then    
        rm -rf $gmapper_index_dir
    fi
fi
echo "... done"
