#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "common.h"
#include "genome.h"

int num_contigs=0 ;

char **contig_fnames=NULL ;
char **contig_fasta_fnames=NULL ;
char **contig_names=NULL ;
char *alph="acgt" ;
FILE **contig_files=NULL ;

int num_est_files=0 ;
char **est_fnames=NULL ;

int num_prot_files=0 ;
char **prot_fnames=NULL ;

int num_cdna_files=0 ;
char **cdna_fnames=NULL ;

int num_annotation_files=0 ;
char **annotation_fnames=NULL ;

char *basedir=NULL ;

// some defaults
int MAX_DOUBLE_QGAP    = 5 ; 
int MAX_DOUBLE_TGAP    = 5 ;
int MAX_CDNA_DELETION  = 3 ;
int MAX_CDNA_INSERTION = 3 ;
int MIN_INTRON_LEN    = 20 ;
int MAX_INTRON_LEN    = 50000 ;
int TERMINAL_EXON_END_TOL = 3 ;
int MERGE_EST_TRANSCRIPTS = 1 ;
int MAX_PERFECT_EST_MATCH_LEN = 100000 ;
int BLAT_BEST_HIT_ONLY = 0 ;
int MIN_SSQUALITY_SCORE = 100 ;
int MIN_EXONQUALITY_SCORE = 95 ;

double BLAT_BEST_HIT_MARGIN = 0.05 ;
double MIN_EST_COVER_FRAC = 0.95 ;
double MIN_PROT_COVER_FRAC = 0.95 ;
double MIN_CDNA_COVER_FRAC = 0.99 ;

char *path_with_basedir(char* b1, char* basedir)
{
	char buf[1000] ;
	if (b1[0]!='/')
	{
		if (basedir[strlen(basedir)-1]=='/')
			sprintf(buf, "%s%s", basedir, b1);
		else
			sprintf(buf, "%s/%s", basedir, b1);
		return strdup(buf) ;
	} else
		return strdup(b1) ;
}

void cleanup_genome() 
{
	free(basedir) ;
	int i;
	for (i=0; i<num_contigs; i++)
	{
		free(contig_fnames[i]) ;
		free(contig_fasta_fnames[i]) ;
		free(contig_names[i]) ;
	}
	free(contig_fnames) ;
	free(contig_fasta_fnames) ;
	free(contig_names) ;

	for (i=0; i<num_contigs; i++)
		if (contig_files[i])
			fclose(contig_files[i]) ;
	free(contig_files) ;
	
	for (i=0; i<num_est_files; i++)
		free(est_fnames[i]) ;
	free(est_fnames) ;

	for (i=0; i<num_prot_files; i++)
		free(prot_fnames[i]) ;
	free(prot_fnames) ;

	for (i=0; i<num_cdna_files; i++)
		free(cdna_fnames[i]) ;
	free(cdna_fnames) ;
	cdna_fnames=NULL ;
	num_cdna_files=0 ;

	for (i=0; i<num_annotation_files; i++)
		free(annotation_fnames[i]) ;
	free(annotation_fnames) ;

	free(alph) ;
}

int init_genome(char *fname)
{
	if (fname==NULL)
		fname="genome.conf" ;
	//fprintf(stderr, "%s\n", fname) ;
	
	FILE * fd=fopen(fname, "r") ;
	if (fd==NULL)
		return -1 ;
	int contigs_found = 0 ;
	int ests_found = 0 ;
	int cdnas_found = 0 ;
	int alph_found=0 ;
	int basedir_found=0 ;
	int annotation_found=0 ;
	basedir = (char*) malloc(1000*sizeof(char)) ;
   
	while (!feof(fd))
	{
		char line[1000];
		char *ret=read_line(fd, line, 1000) ;
		if (ret==NULL)
			break ;
		if (line[0]=='#')
			continue ;
		
		if (strncmp(line,"BASEDIR",7)==0)
		{
			basedir_found = 1 ;
			int args=sscanf(line, "BASEDIR %s", basedir) ;
			if (args!=1)
			{
				fprintf(stderr, "string argument expected after BASEDIR\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MIN_EST_COVER_FRAC", 18)==0)
		{
			int args=sscanf(line, "MIN_EST_COVER_FRAC %lf", &MIN_EST_COVER_FRAC) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MIN_EST_COVER_FRAC\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MIN_PROT_COVER_FRAC", 19)==0)
		{
			int args=sscanf(line, "MIN_PROT_COVER_FRAC %lf", &MIN_PROT_COVER_FRAC) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MIN_PROT_COVER_FRAC\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MIN_CDNA_COVER_FRAC", 19)==0)
		{
			int args=sscanf(line, "MIN_CDNA_COVER_FRAC %lf", &MIN_CDNA_COVER_FRAC) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MIN_CDNA_COVER_FRAC\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MAX_DOUBLE_QGAP", strlen("MAX_DOUBLE_QGAP"))==0)
		{
			int args=sscanf(line, "MAX_DOUBLE_QGAP %i", &MAX_DOUBLE_QGAP) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MAX_DOUBLE_QGAP\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MAX_DOUBLE_TGAP", strlen("MAX_DOUBLE_TGAP"))==0)
		{
			int args=sscanf(line, "MAX_DOUBLE_TGAP %i", &MAX_DOUBLE_TGAP) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MAX_DOUBLE_TGAP\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MAX_CDNA_DELETION", strlen("MAX_CDNA_DELETION"))==0)
		{
			int args=sscanf(line, "MAX_CDNA_DELETION %i", &MAX_CDNA_DELETION) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MAX_CDNA_DELETION\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MAX_CDNA_INSERTION", strlen("MAX_CDNA_INSERTION"))==0)
		{
			int args=sscanf(line, "MAX_CDNA_INSERTION %i", &MAX_CDNA_INSERTION) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MAX_CDNA_INSERTION\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MIN_INTRON_LEN", strlen("MIN_INTRON_LEN"))==0)
		{
			int args=sscanf(line, "MIN_INTRON_LEN %i", &MIN_INTRON_LEN) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after MIN_INTRON_LEN\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MAX_INTRON_LEN", 14)==0)
		{
			int args=sscanf(line, "MAX_INTRON_LEN %i", &MAX_INTRON_LEN) ;
			if (args!=1)
			{
				fprintf(stderr, "integer argument expected after MAX_INTRON_LEN\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"TERMINAL_EXON_END_TOL", strlen("TERMINAL_EXON_END_TOL"))==0)
		{
			int args=sscanf(line, "TERMINAL_EXON_END_TOL %i", &TERMINAL_EXON_END_TOL) ;
			if (args!=1)
			{
				fprintf(stderr, "real argument expected after TERMINAL_EXON_END_TOL\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MERGE_EST_TRANSCRIPTS", strlen("MERGE_EST_TRANSCRIPTS"))==0)
		{
			int args=sscanf(line, "MERGE_EST_TRANSCRIPTS %i", &MERGE_EST_TRANSCRIPTS) ;
			if (args!=1 || ((MERGE_EST_TRANSCRIPTS!=0) && (MERGE_EST_TRANSCRIPTS!=1)))
			{
				fprintf(stderr, "0 or 1 expected after MERGE_EST_TRANSCRIPTS\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MAX_PERFECT_EST_MATCH_LEN", strlen("MAX_PERFECT_EST_MATCH_LEN"))==0)
		{
			int args=sscanf(line, "MAX_PERFECT_EST_MATCH_LEN %i", &MAX_PERFECT_EST_MATCH_LEN) ;
			if (args!=1)
			{
				fprintf(stderr, "integer expected after MAX_PERFECT_EST_MATCH_LEN\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"BLAT_BEST_HIT_ONLY", strlen("BLAT_BEST_HIT_ONLY"))==0)
		{
			int args=sscanf(line, "BLAT_BEST_HIT_ONLY %i", &BLAT_BEST_HIT_ONLY) ;
			if (args!=1 || (BLAT_BEST_HIT_ONLY!=0 && BLAT_BEST_HIT_ONLY!=1))
			{
				fprintf(stderr, "integer expected after BLAT_BEST_HIT_ONLY\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"BLAT_BEST_HIT_MARGIN", strlen("BLAT_BEST_HIT_MARGIN"))==0)
		{
			int args=sscanf(line, "BLAT_BEST_HIT_MARGIN %lf", &BLAT_BEST_HIT_MARGIN) ;
			if (args!=1)
			{
				fprintf(stderr, "integer expected after BLAT_BEST_HIT_MARGIN\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MIN_SSQUALITY_SCORE", strlen("MIN_SSQUALITY_SCORE"))==0)
		{
			int args=sscanf(line, "MIN_SSQUALITY_SCORE %i", &MIN_SSQUALITY_SCORE) ;
			if (args!=1)
			{
				fprintf(stderr, "integer expected after MIN_SSQUALITY_SCORE\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"MIN_EXONQUALITY_SCORE", strlen("MIN_EXONQUALITY_SCORE"))==0)
		{
			int args=sscanf(line, "MIN_EXONQUALITY_SCORE %i", &MIN_EXONQUALITY_SCORE) ;
			if (args!=1)
			{
				fprintf(stderr, "integer expected after MIN_EXONQUALITY_SCORE\n") ;
				fclose(fd) ;
				return -1 ;
			}
		}
		else if (strncmp(line,"CONTIGS",7)==0)
		{
			if (strlen(basedir)==0)
			{
				fprintf(stderr, "BASEDIR line has to appear before all other config lines\n") ;
				fclose(fd) ;
				return -1 ;
			}
			
			contigs_found = 1 ;
			int args=sscanf(line, "CONTIGS %i", &num_contigs) ;
			if (args!=1)
			{
				fprintf(stderr, "integer argument expected after CONTIGS\n") ;
				fclose(fd) ;
				return -1 ;
			}
			//fprintf(stderr, "Using %i contigs\n", num_contigs) ;
			contig_fnames=(char**)malloc(num_contigs*sizeof(char*)) ;
			contig_fasta_fnames=(char**)malloc(num_contigs*sizeof(char*)) ;
			contig_names=(char**)malloc(num_contigs*sizeof(char*)) ;
			contig_files=(FILE**)malloc(num_contigs*sizeof(FILE*)) ;
			
			int i ;
			for (i=0; i<num_contigs; i++)
			{
				contig_fnames[i]=NULL ;
				contig_fasta_fnames[i]=NULL ;
				contig_names[i]=NULL ;
				contig_files[i]=NULL ;
				
				char b1[1000], b2[1000], b3[1000] ;
				char *ret=read_line(fd, line, 1000) ;
				assert(ret!=NULL) ;
				int args=sscanf(line, "%s %s %s", b1, b2, b3) ;
				assert(args==3) ;
				contig_names[i]=strdup(b1) ;
				contig_fnames[i]=path_with_basedir(b2,basedir) ;
				contig_fasta_fnames[i]=path_with_basedir(b3,basedir) ;
				//if (num_contigs<30)
				//fprintf(stderr, "%s in file %s\n", contig_names[i], contig_fnames[i]) ;
			}
		}
		else if (strncmp(line,"ESTFILES",8)==0)
		{
			if (strlen(basedir)==0)
			{
				fprintf(stderr, "BASEDIR line has to appear before all other config lines\n") ;
				fclose(fd) ;
				return -1 ;
			}
			ests_found = 1 ;
			int args=sscanf(line, "ESTFILES %i", &num_est_files) ;
			if (args!=1)
			{
				fprintf(stderr, "integer argument expected after ESTFILES\n") ;
				fclose(fd) ;
				return -1 ;
			}
			est_fnames=(char**)malloc(num_est_files*sizeof(char*)) ;
			
			int i ;
			for (i=0; i<num_est_files; i++)
			{
				est_fnames[i]=NULL ;
				
				char b1[1000] ;
				char *ret=read_line(fd, line, 1000) ;
				assert(ret!=NULL) ;
				int args=sscanf(line, "%s", b1) ;
				assert(args==1) ;
				est_fnames[i]=path_with_basedir(b1,basedir) ;
			}
		}
		else if (strncmp(line,"PROTFILES",9)==0)
		{
			if (strlen(basedir)==0)
			{
				fprintf(stderr, "BASEDIR line has to appear before all other config lines\n") ;
				fclose(fd) ;
				return -1 ;
			}
			int args=sscanf(line, "PROTFILES %i", &num_prot_files) ;
			if (args!=1)
			{
				fprintf(stderr, "integer argument expected after PROTFILES\n") ;
				fclose(fd) ;
				return -1 ;
			}
			prot_fnames=(char**)malloc(num_prot_files*sizeof(char*)) ;
			
			int i ;
			for (i=0; i<num_prot_files; i++)
			{
				prot_fnames[i]=NULL ;
				
				char b1[1000] ;
				char *ret=read_line(fd, line, 1000) ;
				assert(ret!=NULL) ;
				int args=sscanf(line, "%s", b1) ;
				assert(args==1) ;
				prot_fnames[i]=path_with_basedir(b1,basedir) ;
			}
		}
		else if ((strncmp(line,"CDNAFILES",9)==0) || (strncmp(line,"FLCDNAFILES",11)==0))
		{
			if (strlen(basedir)==0)
			{
				fprintf(stderr, "BASEDIR line has to appear before all other config lines\n") ;
				fclose(fd) ;
				return -1 ;
			}
			cdnas_found = 1 ;
			int args = 0 ;
			int old_num_cdna_files = num_cdna_files ;
			if (strncmp(line,"CDNAFILES",9)==0)
				args = sscanf(line, "CDNAFILES %i", &num_cdna_files) ;
			else
				args = sscanf(line, "FLCDNAFILES %i", &num_cdna_files) ;
			if (args!=1)
			{
				fprintf(stderr, "integer argument expected after [FL]CDNAFILES\n") ;
				fclose(fd) ;
				return -1 ;
			}
			if (num_cdna_files>0)
			{
				if (cdna_fnames)
				{
					assert(old_num_cdna_files!=0) ;
					cdna_fnames = (char**)realloc(cdna_fnames, (num_cdna_files+old_num_cdna_files)*sizeof(char*)) ;
				}
				else
					cdna_fnames = (char**)malloc(num_cdna_files*sizeof(char*)) ;
				
				int i ;
				for (i=0; i<num_cdna_files; i++)
				{
					cdna_fnames[i+old_num_cdna_files]=NULL ;
					
					char b1[1000] ;
					char *ret=read_line(fd, line, 1000) ;
					assert(ret!=NULL) ;
					int args=sscanf(line, "%s", b1) ;
					assert(args==1) ;
					cdna_fnames[i+old_num_cdna_files] = path_with_basedir(b1,basedir) ;
				}
				num_cdna_files += old_num_cdna_files ;
			}
		}
		else if (strncmp(line,"ANNOTATIONFILES",15)==0)
		{
			if (strlen(basedir)==0)
			{
				fprintf(stderr, "BASEDIR line has to appear before all other config lines\n") ;
				fclose(fd) ;
				return -1 ;
			}
			annotation_found = 1 ;
			int args=sscanf(line, "ANNOTATIONFILES %i", &num_annotation_files) ;
			if (args!=1)
			{
				fprintf(stderr, "integer argument expected after ANNOTATIONFILES\n") ;
				fclose(fd) ;
				return -1 ;
			}
			annotation_fnames=(char**)malloc(num_annotation_files*sizeof(char*)) ;
			
			int i ;
			for (i=0; i<num_annotation_files; i++)
			{
				annotation_fnames[i]=NULL ;
				
				char b1[1000] ;
				char *ret=read_line(fd, line, 1000) ;
				assert(ret!=NULL) ;
				int args=sscanf(line, "%s", b1) ;
				assert(args==1) ;
				annotation_fnames[i]=path_with_basedir(b1,basedir) ;
			}
		}
		else if (strncmp(line,"ALPHABET",8)==0)
		{
			alph_found=1 ;
			char buf[1000] ;
			int args=sscanf(line, "ALPHABET %s", buf) ;
			assert(args==1) ;
			assert(strlen(buf)==4) ;
			alph=strdup(buf) ;
//			fprintf(stderr, "using alphabet '%s'\n", alph) ;
		}
		else if (strlen(line)>0)
		{
			fprintf(stderr, "unrecognized line '%s'\n", line) ;
			fclose(fd) ;
			return -1 ;
		}
	}
	
	assert(alph_found==1) ;
	assert(contigs_found==1) ;
	assert(ests_found==1) ;
	assert(cdnas_found==1) ;
	assert(basedir_found==1) ;
	assert(annotation_found==1) ;

	fclose(fd) ;
	return 0 ;
}

static char last_contig_name[100]="" ;
static int last_contig_num=0 ;

/* returns the contig idx from the string */
int get_contig_num(char *contig_name)
{
	int i, contig_num=-1 ;
	// when there are enough contigs, it is worth caching 
	// the last result
	if ((num_contigs>10) && (strcmp(contig_name, last_contig_name)==0))
		return last_contig_num ;
	
	for (i=0; i<num_contigs; i++)
		if (strcmp(contig_name, contig_names[i])==0)
			contig_num=i ;
	if (contig_num==-1)
    {
		fprintf(stderr, "contig name not found (%s)\n", contig_name) ;
		exit(-1) ;
    } ;
	if (num_contigs>10)
	{
		last_contig_num = contig_num ;
		strncpy(last_contig_name, contig_name, 99) ;
		last_contig_name[99]=0 ;
	} ;
	return contig_num ;
} 
